# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from collections import namedtuple
from typing import (
    Any,
    Callable,
    Optional,
    Union,
)

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.node_data import Pwc
from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
    nullable,
)
from boulderopal.graph import (
    Graph,
    execute_graph,
)
from boulderopal.optimization import run_optimization
from boulderopal.superconducting._components import (
    Cavity,
    CavityCavityInteraction,
    Coefficient,
    Transmon,
    TransmonCavityInteraction,
    TransmonTransmonInteraction,
    validate_real_coefficient,
)
from boulderopal.superconducting._drives import (
    ComplexOptimizableConstant,
    ComplexOptimizableSignal,
    RealOptimizableConstant,
    RealOptimizableSignal,
)

Interaction = Union[
    TransmonTransmonInteraction, TransmonCavityInteraction, CavityCavityInteraction
]

_QHOOps = namedtuple("_QHOOps", ["a", "adag", "n"])


def _create_qho_operators(
    graph: Graph, transmons: list[Transmon], cavities: list[Cavity]
) -> dict[str, _QHOOps]:
    """
    Create the creation, annihilation, and number operators for transmons and cavities.

    This function returns a dictionary whose keys are the subsystem names and whose values are
    _QHOOps named tuples (with field names a, adag, and n) with the operators of each subsystem.
    """
    systems = transmons + cavities
    dimensions = [system.dimension for system in systems]
    return {
        system.name: _QHOOps(
            graph.embed_operators(
                (graph.annihilation_operator(system.dimension), idx), dimensions
            ),
            graph.embed_operators(
                (graph.creation_operator(system.dimension), idx), dimensions
            ),
            graph.embed_operators(
                (graph.number_operator(system.dimension), idx), dimensions
            ),
        )
        for idx, system in enumerate(systems)
    }


def _validate_physical_system_inputs(
    transmons: list[Transmon], cavities: list[Cavity], interactions: list[Interaction]
) -> tuple[
    list[TransmonTransmonInteraction],
    list[TransmonCavityInteraction],
    list[CavityCavityInteraction],
]:
    """
    Validate the type of the subsystems and interactions,
    and the subsystem names in the interactions.

    This function returns three lists containing
    (1) all the transmon-transmon interactions,
    (2) all the transmon-cavity interactions, and
    (3) all the cavity-cavity interactions.
    """

    check_argument(
        len(transmons) + len(cavities) > 0,
        "At least one transmon or cavity must be provided.",
        {"transmons": transmons, "cavities": cavities},
    )

    check_argument(
        all(isinstance(transmon, Transmon) for transmon in transmons),
        "Each element in transmons must be a Transmon object.",
        {"transmons": transmons},
    )
    transmon_names = [transmon.name for transmon in transmons]
    check_argument(
        len(transmon_names) == len(set(transmon_names)),
        "Transmon names must be unique.",
        {"transmons": transmons},
        extras={"[transmon.name for transmon in transmons]": transmon_names},
    )

    check_argument(
        all(isinstance(cavity, Cavity) for cavity in cavities),
        "Each element in cavities must be a Cavity object.",
        {"cavities": cavities},
    )
    cavity_names = [cavity.name for cavity in cavities]
    check_argument(
        len(cavity_names) == len(set(cavity_names)),
        "Cavity names must be unique.",
        {"cavities": cavities},
        extras={"[cavity.name for cavity in cavities]": cavity_names},
    )

    check_argument(
        len(transmon_names + cavity_names) == len(set(transmon_names + cavity_names)),
        "Transmon and cavity names must be unique.",
        {"transmons": transmons, "cavities": cavities},
        extras={
            "[transmon.name for transmon in transmons]": transmon_names,
            "[cavity.name for cavity in cavities]": cavity_names,
        },
    )

    transmon_transmon_interactions = [
        intx for intx in interactions if isinstance(intx, TransmonTransmonInteraction)
    ]
    transmon_cavity_interactions = [
        intx for intx in interactions if isinstance(intx, TransmonCavityInteraction)
    ]
    cavity_cavity_interactions = [
        intx for intx in interactions if isinstance(intx, CavityCavityInteraction)
    ]
    check_argument(
        len(transmon_transmon_interactions)
        + len(transmon_cavity_interactions)
        + len(cavity_cavity_interactions)
        == len(interactions),
        "Each element in interactions must be a TransmonTransmonInteraction, "
        "a TransmonCavityInteraction, or a CavityCavityInteraction object.",
        {"interactions": interactions},
    )

    tt_interaction_name_pairs = set(
        frozenset(tt_interaction.transmon_names)
        for tt_interaction in transmon_transmon_interactions
    )
    check_argument(
        len(tt_interaction_name_pairs) == len(transmon_transmon_interactions),
        "There are duplicate names in the transmon-transmon interaction terms.",
        {"interactions": interactions},
    )

    for tt_interaction in transmon_transmon_interactions:
        name_1, name_2 = tt_interaction.transmon_names
        check_argument(
            name_1 in transmon_names and name_2 in transmon_names,
            "Names in transmon-transmon interaction terms must refer to transmons in the system.",
            {"transmons": transmons, "interactions": interactions},
            extras={"[transmon.name for transmon in transmons]": transmon_names},
        )

    tc_interaction_name_pairs = set(
        (tc_interaction.transmon_name, tc_interaction.cavity_name)
        for tc_interaction in transmon_cavity_interactions
    )
    check_argument(
        len(tc_interaction_name_pairs) == len(transmon_cavity_interactions),
        "There are duplicate names in the transmon-cavity interaction terms.",
        {"interactions": interactions},
    )

    for tc_interaction in transmon_cavity_interactions:
        check_argument(
            tc_interaction.transmon_name in transmon_names,
            "Transmon names in transmon-cavity interaction terms "
            "must refer to transmons in the system.",
            {"transmons": transmons, "interactions": interactions},
            extras={"[transmon.name for transmon in transmons]": transmon_names},
        )
        check_argument(
            tc_interaction.cavity_name in cavity_names,
            "Cavity names in transmon-cavity interaction terms "
            "must refer to cavities in the system.",
            {"cavities": cavities, "interactions": interactions},
            extras={"[cavity.name for cavity in cavities]": cavity_names},
        )

    cc_interaction_name_pairs = set(
        frozenset(cc_interaction.cavity_names)
        for cc_interaction in cavity_cavity_interactions
    )
    check_argument(
        len(cc_interaction_name_pairs) == len(cavity_cavity_interactions),
        "There are duplicate names in the cavity-cavity interaction terms.",
        {"interactions": interactions},
    )

    for cc_interaction in cavity_cavity_interactions:
        name_1, name_2 = cc_interaction.cavity_names
        check_argument(
            name_1 in cavity_names and name_2 in cavity_names,
            "Names in cavity-cavity interaction terms must refer to cavities in the system.",
            {"cavities": cavities, "interactions": interactions},
            extras={"[cavity.name for cavity in cavities]": cavity_names},
        )

    return (
        transmon_transmon_interactions,
        transmon_cavity_interactions,
        cavity_cavity_interactions,
    )


def _convert_to_pwc(
    graph: Graph,
    coefficient: Coefficient,
    real_valued: bool,
    name: str,
    filter_signal: Optional[Callable],
    gate_duration: float,
    optimizable_node_names: list,
) -> Pwc:
    """
    Return the Pwc representation of a coefficient.

    This function has a side effect to mutate `optimizable_node_names` list as
    it iterates over the coefficient parameter.
    """

    # This is to double check we pass the real coefficient when needed.
    # Also note this is done by checking whether the corresponding Hamiltonian part
    # is Hermitian or not, which is due to the format of the Hamiltonian and the process of how
    # we build the Hamiltonian from user input.
    if real_valued:
        coefficient = validate_real_coefficient(coefficient, name)

    # Convert scalar value into constant Pwc.
    if np.isscalar(coefficient):
        return graph.constant_pwc(
            constant=graph.tensor(coefficient, name=name), duration=gate_duration
        )

    # Convert array into Pwc.
    if isinstance(coefficient, np.ndarray):
        return graph.pwc_signal(coefficient, gate_duration, name=name)

    # Convert Real/ComplexOptimizableConstant into optimizable constant Pwc.
    if isinstance(coefficient, (RealOptimizableConstant, ComplexOptimizableConstant)):
        optimizable_node_names.append(name)
        return coefficient.get_pwc(graph, gate_duration, name)

    # Convert Real/ComplexOptimizableSignal into optimizable Pwc.
    if isinstance(coefficient, (RealOptimizableSignal, ComplexOptimizableSignal)):
        optimizable_node_names.append(name)
        signal = coefficient.get_pwc(graph, gate_duration, name)
        if filter_signal is None:
            return signal
        optimizable_node_names.append(f"{name}_filtered")
        return filter_signal(signal, name)

    raise QctrlArgumentsValueError(f"{name} has an invalid type.", {name: coefficient})


def _create_superconducting_hamiltonian(
    graph: Graph,
    transmons: list[Transmon],
    cavities: list[Cavity],
    interactions: list[Interaction],
    gate_duration: float,
    cutoff_frequency: Optional[float],
    sample_count: int,
) -> tuple[Pwc, list[str]]:
    """
    Create the Hamiltonian of a system composed of transmons and cavities.

    This function returns the Hamiltonian as a Pwc node and a list with the names of
    the optimizable nodes that have been added to the graph. If some of these are PWC functions
    and `cutoff_frequency` is not None, then the names of the filtered PWC nodes are also included.
    """

    (
        transmon_transmon_interactions,
        transmon_cavity_interactions,
        cavity_cavity_interactions,
    ) = _validate_physical_system_inputs(transmons, cavities, interactions)

    # Define annihilation and creation operators for the transmon and the cavity.
    operators = _create_qho_operators(graph, transmons, cavities)

    # Create nested dictionary structure containing information for the different Hamiltonian terms.
    hamiltonian_info: dict[str, Any] = {}

    # Add transmon terms.
    for transmon in transmons:
        transmon_ops = operators[transmon.name]
        hamiltonian_info[f"{transmon.name}.frequency"] = {
            "coefficient": transmon.frequency,
            "operator": transmon_ops.n,
            "is_hermitian": True,
        }
        hamiltonian_info[f"{transmon.name}.anharmonicity"] = {
            "coefficient": transmon.anharmonicity,
            "operator": 0.5 * (transmon_ops.n @ transmon_ops.n - transmon_ops.n),
            "is_hermitian": True,
        }
        hamiltonian_info[f"{transmon.name}.drive"] = {
            "coefficient": transmon.drive,
            "operator": transmon_ops.adag,
            "is_hermitian": False,
        }

    # Add cavity terms.
    for cavity in cavities:
        cavity_ops = operators[cavity.name]
        hamiltonian_info[f"{cavity.name}.frequency"] = {
            "coefficient": cavity.frequency,
            "operator": cavity_ops.n,
            "is_hermitian": True,
        }
        hamiltonian_info[f"{cavity.name}.kerr_coefficient"] = {
            "coefficient": cavity.kerr_coefficient,
            "operator": 0.5 * (cavity_ops.n @ cavity_ops.n - cavity_ops.n),
            "is_hermitian": True,
        }
        hamiltonian_info[f"{cavity.name}.drive"] = {
            "coefficient": cavity.drive,
            "operator": cavity_ops.adag,
            "is_hermitian": False,
        }

    # Add transmon-transmon interaction terms.
    for tt_interaction in transmon_transmon_interactions:
        name_1, name_2 = tt_interaction.transmon_names
        transmon_1_ops = operators[name_1]
        transmon_2_ops = operators[name_2]
        key = f"{name_1}_{name_2}_interaction"
        hamiltonian_info[f"{key}.effective_coupling"] = {
            "coefficient": tt_interaction.effective_coupling,
            "operator": 2 * transmon_1_ops.a @ transmon_2_ops.adag,
            "is_hermitian": False,
        }

    # Add transmon-cavity interaction terms.
    for tc_interaction in transmon_cavity_interactions:
        transmon_ops = operators[tc_interaction.transmon_name]
        cavity_ops = operators[tc_interaction.cavity_name]
        key = f"{tc_interaction.transmon_name}_{tc_interaction.cavity_name}_interaction"
        hamiltonian_info[f"{key}.dispersive_shift"] = {
            "coefficient": tc_interaction.dispersive_shift,
            "operator": transmon_ops.n @ cavity_ops.n,
            "is_hermitian": True,
        }
        hamiltonian_info[f"{key}.rabi_coupling"] = {
            "coefficient": tc_interaction.rabi_coupling,
            "operator": 2 * transmon_ops.adag @ cavity_ops.a,
            "is_hermitian": False,
        }

    # Add cavity-cavity interaction terms.
    for cc_interaction in cavity_cavity_interactions:
        name_1, name_2 = cc_interaction.cavity_names
        cavity_1_ops = operators[name_1]
        cavity_2_ops = operators[name_2]
        key = f"{name_1}_{name_2}_interaction"
        hamiltonian_info[f"{key}.cross_kerr_coefficient"] = {
            "coefficient": cc_interaction.cross_kerr_coefficient,
            "operator": cavity_1_ops.n @ cavity_2_ops.n,
            "is_hermitian": True,
        }

    check_argument(
        any(np.any(info["coefficient"]) for info in hamiltonian_info.values()),
        "The system must contain at least one Hamiltonian coefficient.",
        {"transmons": transmons, "cavities": cavities, "interactions": interactions},
    )

    # Create kernel to filter signals (used in convert_to_pwc).
    if cutoff_frequency is not None:
        kernel = graph.sinc_convolution_kernel(cutoff_frequency)
        _filter = lambda x, y: graph.discretize_stf(  # pylint: disable=unnecessary-lambda-assignment
            stf=graph.convolve_pwc(pwc=x, kernel=kernel),
            duration=gate_duration,
            segment_count=sample_count,
            name=f"{y}_filtered",
        )
    else:
        _filter = None

    # Build the Hamiltonian from the different terms.
    hamiltonian_terms = []
    optimizable_node_names: list[str] = []  # filled up by convert_to_pwc

    for name, info in hamiltonian_info.items():
        if info["coefficient"] is not None:
            coefficient = _convert_to_pwc(
                graph=graph,
                coefficient=info["coefficient"],
                real_valued=info["is_hermitian"],
                name=name,
                filter_signal=_filter,
                gate_duration=gate_duration,
                optimizable_node_names=optimizable_node_names,
            )
            if info["is_hermitian"]:
                hamiltonian_terms.append(coefficient * info["operator"])
            else:
                hamiltonian_terms.append(0.5 * coefficient * info["operator"])
                hamiltonian_terms.append(
                    0.5 * graph.conjugate(coefficient) * graph.adjoint(info["operator"])
                )

    return graph.pwc_sum(hamiltonian_terms), optimizable_node_names


def simulate(
    transmons: list[Transmon],
    cavities: list[Cavity],
    interactions: list[Interaction],
    gate_duration: float,
    sample_count: int = 128,
    cutoff_frequency: Optional[float] = None,
    initial_state: Optional[np.ndarray] = None,
) -> dict:
    """
    Simulate a system composed of transmons and cavities.

    This function builds a graph describing the Hamiltonian (see the note part for details) of a
    superconducting system, and calls :func:`~boulderopal.execute_graph` to simulate its dynamics.

    Parameters
    ----------
    transmons : list[Transmon]
        List of objects containing the physical information about the transmons.
        It must not contain any optimizable coefficients.
        It can be an empty list, but at least one transmon or cavity must be provided.
    cavities : list[Cavity]
        List of objects containing the physical information about the cavities.
        They must not contain any optimizable coefficients.
        It can be an empty list, but at least one transmon or cavity must be provided.
    interactions : list[TransmonTransmonInteraction or TransmonCavityInteraction or \
            CavityCavityInteraction]
        List of objects containing the physical information about the interactions in the system.
        They must not contain any optimizable coefficients.
        It can be an empty list.
    gate_duration : float
        The duration of the gate to be simulated, :math:`t_\\mathrm{gate}`.
        It must be greater than zero.
    sample_count : int, optional
        The number of times between 0 and `gate_duration` (included)
        at which the evolution is sampled.
        Defaults to 128.
    cutoff_frequency : float or None, optional
        The cutoff frequency of a linear sinc filter to be applied to the piecewise-constant
        signals you provide for the coefficients. If not provided, the signals are not filtered.
        If the signals are filtered, a larger sample count leads to a more accurate numerical
        integration. If the signals are not filtered, the sample count has no effect on the
        numerical precision of the integration.
    initial_state : np.ndarray or None, optional
        The initial state of the system, :math:`|\\Psi_\\mathrm{initial}\\rangle`, as a 1D array of
        length ``D = np.prod([system.dimension for system in transmons + cavities])``.
        If not provided, the function only returns the system's unitary time-evolution operators.

    Returns
    -------
    dict
        The result of the `execute_graph` call.
        Its ``output`` item is a dictionary containing information about
        the evolution of the system, with the following keys:

            ``sample_times``
                The times at which the system's evolution is sampled,
                as an array of shape ``(T,)``.
            ``unitaries``
                The system's unitary time-evolution operators at each sample time,
                as an array of shape ``(T, D, D)``.
            ``state_evolution``
                The time evolution of the initial state at each sample time,
                as an array of shape ``(T, D)``.
                This is only returned if you provide an initial state.

    See Also
    --------
    :func:`boulderopal.superconducting.optimize` :
        Find optimal pulses or parameters for a system composed of transmons and cavities.

    Notes
    -----
    The Hamiltonian of the system is of the form

    .. math::
        H = \\sum_i H_{\\mathrm{transmon}_i}
            + \\sum_i H_{\\mathrm{cavity}_i}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{transmon}_j}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{cavity}_j}
            + \\sum_{i,j} H_{\\mathrm{cavity}_i-\\mathrm{cavity}_j}

    where i and j mark the i-th and j-th transmon or cavity.
    For their definition of each Hamiltonian term, see its respective class.

    The Hilbert space of the system is defined as the outer product of all the
    transmon Hilbert spaces (in the order they're provided in `transmons`) with
    the cavity Hilbert spaces (in the order they're provided in `cavities`), that is:

    .. math::
        \\mathcal{H} =
            \\mathcal{H}_{\\mathrm{transmon}_1} \\otimes \\mathcal{H}_{\\mathrm{transmon}_2}
            \\otimes \\ldots
            \\otimes \\mathcal{H}_{\\mathrm{cavity}_1} \\otimes \\mathcal{H}_{\\mathrm{cavity}_2}
            \\otimes \\ldots

    The system dimension `D` is then the product of all transmon and cavity dimensions.
    """
    gate_duration = ScalarDType.REAL(gate_duration, min_=0, name="gate_duration")
    sample_count = ScalarDType.INT(sample_count, min_=0, name="sample_count")
    cutoff_frequency = nullable(
        ScalarDType.REAL, cutoff_frequency, name="cutoff_frequency"
    )

    system_dimension = np.prod(
        [system.dimension for system in transmons + cavities], dtype=int
    )
    initial_state = nullable(
        ArrayDType.COMPLEX,
        initial_state,
        ndim=1,
        shape=(system_dimension,),
        name="initial_state",
    )

    graph = Graph()

    # Create PWC Hamiltonian.
    hamiltonian, optimizable_node_names = _create_superconducting_hamiltonian(
        graph=graph,
        transmons=transmons,
        cavities=cavities,
        interactions=interactions,
        gate_duration=gate_duration,
        cutoff_frequency=cutoff_frequency,
        sample_count=sample_count,
    )

    # Check whether there are any optimizable coefficients.
    check_argument(
        len(optimizable_node_names) == 0,
        "None of the Hamiltonian terms can be optimizable.",
        {"transmons": transmons, "cavities": cavities, "interactions": interactions},
    )

    # Calculate the evolution.
    sample_times = np.linspace(0.0, gate_duration, sample_count)
    graph.tensor(sample_times, name="sample_times")
    unitaries = graph.time_evolution_operators_pwc(
        hamiltonian=hamiltonian, sample_times=sample_times, name="unitaries"
    )

    output_node_names = ["unitaries", "sample_times"]

    if initial_state is not None:
        states = unitaries @ initial_state[:, None]
        states = states[..., 0]
        states.name = "state_evolution"
        output_node_names.append("state_evolution")

    return execute_graph(graph=graph, output_node_names=output_node_names)


def optimize(
    transmons: list[Transmon],
    cavities: list[Cavity],
    interactions: list[Interaction],
    gate_duration: float,
    initial_state: Optional[np.ndarray] = None,
    target_state: Optional[np.ndarray] = None,
    target_operation: Optional[np.ndarray] = None,
    sample_count: int = 128,
    cutoff_frequency: Optional[float] = None,
    **optimization_kwargs: Any,
) -> dict[str, Any]:
    """
    Find optimal pulses or parameters for a system composed of transmons and cavities,
    in order to achieve a target state or implement a target operation.

    At least one of the terms in the `transmons`, `cavities`, or `interactions` arguments
    must be optimizable.

    To optimize a state transfer, you need to provide an initial and a target state.
    To optimize a target gate/unitary, you need to provide a target operation.

    This function builds a graph describing the Hamiltonian (see the note part for details) of a
    superconducting system, and calls :func:`~boulderopal.run_optimization` to
    to perform the optimization task.

    Parameters
    ----------
    transmons : list[Transmon]
        List of objects containing the physical information about the transmons.
        It can be an empty list, but at least one transmon or cavity must be provided.
    cavities : list[Cavity]
        List of objects containing the physical information about the cavities.
        It can be an empty list, but at least one transmon or cavity must be provided.
    interactions : list[TransmonTransmonInteraction or TransmonCavityInteraction or \
            CavityCavityInteraction]
        List of objects containing the physical information about the interactions in the system.
        It can be an empty list.
    gate_duration : float
        The duration of the gate to be optimized, :math:`t_\\mathrm{gate}`.
        It must be greater than zero.
    initial_state : np.ndarray or None, optional
        The initial state of the system, :math:`|\\Psi_\\mathrm{initial}\\rangle`, as a 1D array of
        length ``D = np.prod([system.dimension for system in transmon + cavities])``.
        If provided, the function also returns its time evolution.
        This is a required parameter if you pass a `target_state`.
    target_state : np.ndarray or None, optional
        The target state of the optimization, :math:`|\\Psi_\\mathrm{target}\\rangle`,
        as a 1D array of length `D`.
        You must provide exactly one of `target_state` or `target_operation`.
    target_operation : np.ndarray or None, optional
        The target operation of the optimization, :math:`U_\\mathrm{target}`,
        as a 2D array of shape ``(D, D)``.
        You must provide exactly one of `target_state` or `target_operation`.
    sample_count : int, optional
        The number of times between 0 and `gate_duration` (included)
        at which the evolution is sampled.
        Defaults to 128.
    cutoff_frequency : float or None, optional
        The cutoff frequency of a linear sinc filter to be applied to the piecewise-constant
        signals you provide for the coefficients. If not provided, the signals are not filtered.
        If the signals are filtered, a larger sample count leads to a more accurate numerical
        integration. If the signals are not filtered, the sample count has no effect on the
        numerical precision of the integration.
    **optimization_kwargs : dict
        Additional parameters to pass to boulderopal.run_optimization.

    Returns
    -------
    dict
        The result of the `run_optimization` call.
        Its ``output`` item is a dictionary containing the optimized coefficients and information
        about the time evolution of the system, with the following keys:

            optimized coefficients
                These are the names of the requested optimized Hamiltonian coefficients, under
                keys such as ``[transmon_1_name].drive``, ``[cavity_2_name].frequency``,
                ``[transmon_2_name]_[cavity_1_name]_interaction.dispersive_shift``, and
                ``[cavity_1_name]_[cavity_2_name]_interaction.cross_kerr_coefficient`` (where
                ``[transmon_n_name]`` and  ``[cavity_n_name]`` are the names assigned to the
                respective transmons or cavities).
                If you pass a `cutoff_frequency`, the filtered versions of the
                piecewise-constant coefficients are also included with keys such as
                ``[transmon_2_name].drive_filtered``.
            ``infidelity``
                The state/operational infidelity of the optimized evolution.
            ``sample_times``
                The times at which the system's evolution is sampled,
                as an array of shape ``(T,)``.
            ``unitaries``
                The system's unitary time-evolution operators at each sample time,
                as an array of shape ``(T, D, D)``.
            ``state_evolution``
                The time evolution of the initial state at each sample time,
                as an array of shape ``(T, D)``.
                This is only returned if you provide an initial state.

    See Also
    --------
    :func:`boulderopal.superconducting.simulate` :
        Simulate a system composed of transmons and cavities.

    Notes
    -----
    The Hamiltonian of the system is of the form

    .. math::
        H = \\sum_i H_{\\mathrm{transmon}_i}
            + \\sum_i H_{\\mathrm{cavity}_i}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{transmon}_j}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{cavity}_j}
            + \\sum_{i,j} H_{\\mathrm{cavity}_i-\\mathrm{cavity}_j}

    where i and j mark the i-th and j-th transmon or cavity.
    For their definition of each Hamiltonian term, see its respective class.

    The Hilbert space of the system is defined as the outer product of all the
    transmon Hilbert spaces (in the order they're provided in `transmons`) with
    the cavity Hilbert spaces (in the order they're provided in `cavities`), that is:

    .. math::
        \\mathcal{H} =
            \\mathcal{H}_{\\mathrm{transmon}_1} \\otimes \\mathcal{H}_{\\mathrm{transmon}_2}
            \\otimes \\ldots
            \\otimes \\mathcal{H}_{\\mathrm{cavity}_1} \\otimes \\mathcal{H}_{\\mathrm{cavity}_2}
            \\otimes \\ldots

    The system dimension `D` is then the product of all transmon and cavity dimensions.

    If you provide an `initial_state` and a `target_state`, the optimization cost is defined as the
    infidelity of the state transfer process,

    .. math::
        \\mathcal{I}
            = 1 - \\left|
                \\langle
                    \\Psi_\\mathrm{target} | U(t_\\mathrm{gate}) | \\Psi_\\mathrm{initial}
                \\rangle
            \\right|^2 ,

    where :math:`U(t)` is the unitary time-evolution operator generated by the Hamiltonian.

    If you provide a `target_operation`, the optimization cost is defined as the operational
    infidelity,

    .. math::
        \\mathcal{I}
            = 1 - \\left| \\frac
                {\\mathrm{Tr} (U_\\mathrm{target}^\\dagger U(t_\\mathrm{gate}))}
                {\\mathrm{Tr} (U_\\mathrm{target}^\\dagger U_\\mathrm{target})}
            \\right|^2 .

    See the `superconducting systems namespace classes
    <https://docs.q-ctrl.com/boulder-opal/references/boulder-opal/
    boulderopal/superconducting.html>`_
    for a list of the relevant objects to describe subsystems and optimizable coefficients.
    """

    check_argument(
        (target_state is None) ^ (target_operation is None),
        "You have to provide exactly one of `target_state` or `target_operation`.",
        {"target_state": target_state, "target_operation": target_operation},
    )
    if target_state is not None:
        check_argument(
            initial_state is not None,
            "If you provide a `target_state`, you must provide an `initial_state`.",
            {"target_state": target_state, "initial_state": initial_state},
        )

    gate_duration = ScalarDType.REAL(gate_duration, min_=0, name="gate_duration")
    sample_count = ScalarDType.INT(sample_count, min_=0, name="sample_count")
    cutoff_frequency = nullable(
        ScalarDType.REAL, cutoff_frequency, name="cutoff_frequency"
    )

    dim = np.prod([system.dimension for system in transmons + cavities], dtype=int)
    initial_state = nullable(
        ArrayDType.COMPLEX, initial_state, "initial_state", 1, shape=(dim,)
    )
    target_state = nullable(
        ArrayDType.COMPLEX, target_state, "target_state", 1, shape=(dim,)
    )
    target_operation = nullable(
        ArrayDType.COMPLEX, target_operation, "target_operation", 2, shape=(dim, dim)
    )

    graph = Graph()

    # Create PWC Hamiltonian.
    hamiltonian, optimizable_node_names = _create_superconducting_hamiltonian(
        graph=graph,
        transmons=transmons,
        cavities=cavities,
        interactions=interactions,
        gate_duration=gate_duration,
        cutoff_frequency=cutoff_frequency,
        sample_count=sample_count,
    )

    # Check whether there are any optimizable coefficients.
    check_argument(
        len(optimizable_node_names) > 0,
        "At least one of the Hamiltonian terms must be optimizable.",
        {"transmons": transmons, "cavities": cavities, "interactions": interactions},
    )

    sample_times = np.linspace(0.0, gate_duration, sample_count)
    graph.tensor(sample_times, name="sample_times")
    unitaries = graph.time_evolution_operators_pwc(
        hamiltonian=hamiltonian, sample_times=sample_times, name="unitaries"
    )
    other_output_node_names = ["sample_times", "unitaries", "infidelity"]

    if initial_state is not None:
        states = unitaries @ initial_state[:, None]
        states = states[..., 0]
        states.name = "state_evolution"
        other_output_node_names.append("state_evolution")

    if target_state is not None:
        graph.state_infidelity(target_state, states[-1], name="infidelity")
    else:
        graph.unitary_infidelity(unitaries[-1], target_operation, name="infidelity")

    return run_optimization(
        graph=graph,
        cost_node_name="infidelity",
        output_node_names=optimizable_node_names + other_output_node_names,
        **optimization_kwargs,
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Module for nodes to calculate differentiations.
"""
from typing import List

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_iterable,
)

from . import node_data
from .documentation import Category
from .utils import validate_shape


class Hessian(Node):
    r"""
    Calculate a single Hessian matrix for all the variables.

    The Hessian is a matrix containing all the second partial derivatives
    of the `tensor` with respect to the `variables`.

    Parameters
    ----------
    tensor : Tensor(scalar, real)
        The real scalar tensor :math:`T` whose Hessian matrix you want to
        calculate.
    variables : list[Tensor(real)]
        The list of real variables :math:`\{\theta_i\}` with respect to
        which you want to take the second partial derivatives of the
        tensor. If any of the tensors of the list is not scalar, this
        function treats each of the elements of the tensor as a different
        variable. It does this by flattening all tensors and concatenating
        them in the same sequence that you provided in this list.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor(2D, real)
        The real Hessian matrix :math:`H` containing the second partial
        derivatives of the `tensor` :math:`T` with respect to the
        `variables` :math:`\{\theta_i\}`.

    Warnings
    --------
    This function currently doesn't support calculating a Hessian matrix for
    a graph which includes an `infidelity_pwc` node if it involves a Hamiltonian
    with degenerate eigenvalues at any segment. In that case, the function
    returns a `nan` Hessian matrix.

    Notes
    -----
    The :math:`(i,j)` element of the Hessian contains the partial
    derivative of the `tensor` with respect to the ith and the jth
    `variables`:

    .. math::
        H_{i,j} = \frac{\partial^2 T}{\partial \theta_i \partial \theta_j}.

    The variables :math:`\{\theta_i\}` follow the same sequence as the
    input list of `variables`. If some of the `variables` are not scalars,
    this function flattens them and concatenates them in the same order of
    the list of `variables` that you provided to create the sequence of
    scalar variables :math:`\{\theta_i\}`.

    If you choose a negative log-likelihood as the tensor :math:`T`, you
    can use this Hessian as an estimate of the Fisher information matrix.
    """
    name = "hessian"
    args = [
        forge.arg("tensor", type=node_data.Tensor),
        forge.arg("variables", type=List[node_data.Tensor]),
    ]
    rtype = node_data.Tensor
    categories = [Category.LINEAR_ALGEBRA]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        tensor = kwargs.get("tensor")
        variables = kwargs.get("variables")
        check_argument(
            isinstance(tensor, node_data.Tensor),
            "The tensor parameter must be a Tensor.",
            {"tensor": tensor},
        )
        tensor_shape = validate_shape(tensor, "tensor")
        check_argument(
            not tensor_shape, "The tensor must be a scalar tensor.", {"tensor": tensor}
        )
        check_argument_iterable(variables, "variables")
        check_argument(
            all(isinstance(variable, node_data.Tensor) for variable in variables),
            "Each of the variables must be a Tensor.",
            {"variables": variables},
        )
        variable_count = sum(
            np.prod(validate_shape(variable, f"variables[{n}]"), dtype=int)
            for n, variable in enumerate(variables)
        )
        shape = (variable_count, variable_count)
        return node_data.Tensor(_operation, shape=shape)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Module to host some utils functions, which are exposed at the top-level.
"""
from __future__ import annotations

import datetime
import os
from typing import Optional

from qctrlworkflowclient.utils import package_versions_table


def _dict_to_bib(cite_type: str, name: str, data: dict) -> str:
    """
    Convert data dictionary to BibTex format.
    """
    as_list = [f"{key} = {value}" for key, value in data.items()]
    format_string = ",\n  ".join(as_list)
    return f"@{cite_type}{{{name},\n  {format_string}\n}}"


def cite(path: Optional[str] = None) -> None:
    """
    Print the BibTeX information for citing Boulder Opal,
    with the possibility to save it into a BibTeX file.

    Parameters
    ----------
    path : str or None, optional
        If passed, the BibTeX information will be saved to the file 'boulder_opal.bib' at the
        given path.
    """
    paper = cite_boulder_opal_paper()
    print(paper, end="\n\n")
    website = cite_boulder_opal_product()
    print(website)

    if path is not None:
        with open(
            os.path.join(path, "boulder_opal.bib"), "w", encoding="utf-8"
        ) as bibtex_file:
            bibtex_file.write(paper + "\n\n")
            bibtex_file.write(website)


def cite_boulder_opal_paper() -> str:
    """
    Return information for citing the Boulder Opal paper.

    Returns
    -------
    str
        The paper reference.
    """

    data = {
        "doi": "{10.1088/2058-9565/abdca6}",
        "url": "{https://doi.org/10.1088/2058-9565/abdca6}",
        "year": "{2021}",
        "publisher": "{I{OP} Publishing}",
        "volume": "{6}",
        "number": "{4}",
        "pages": "{044011}",
        "author": "{Harrison Ball and Michael J Biercuk and Andre R R Carvalho and Jiayin Chen and "
        "Michael Hush and Leonardo A De Castro and Li Li and Per J Liebermann and Harry "
        "J Slatyer and Claire Edmunds and Virginia Frey and Cornelius Hempel "
        "and Alistair Milne}",
        "title": "{Software tools for quantum control: improving quantum computer performance "
        "through noise and error suppression}",
        "journal": "{Quantum Science and Technology}",
    }
    return _dict_to_bib("article", "boulder_opal1", data)


def cite_boulder_opal_product() -> str:
    """
    Return information for citing the Boulder Opal product (website).

    Returns
    -------
    str
        The website reference.
    """

    data = {
        "author": "{Q-CTRL}",
        "title": "{Boulder {O}pal}",
        "year": f"{{{datetime.datetime.now().year}}}",
        "howpublished": "{https://q-ctrl.com/boulder-opal}",
        "note": "{[Online]}",
    }
    return _dict_to_bib("misc", "boulder_opal2", data)


def print_package_versions() -> None:
    """
    Print a Markdown-formatted table showing the Python version being used,
    as well as the versions of some loaded packages that are relevant to Boulder Opal.
    """

    package_names = [
        # External packages.
        "jsonpickle",
        "matplotlib",
        "mloop",
        "numpy",
        "qiskit",
        "qutip",
        "scipy",
        # Q-CTRL packages.
        "boulderopal",
        "qctrlcommons",
        "qctrlexperimentscheduler",
        "qctrlmloop",
        "qctrlopencontrols",
        "qctrlqua",
        "qctrlvisualizer",
        "qctrlworkflowclient",
    ]

    print(package_versions_table(package_names))


__all__ = ["cite", "print_package_versions"]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import Optional

import numpy as np

from boulderopal._nodes.documentation import Category
from boulderopal._nodes.node_data import (
    ConvolutionKernel,
    Pwc,
)

from .namespace import composite_node


@composite_node([Category.FILTERING_AND_DISCRETIZING])
def filter_and_resample_pwc(
    graph: "Graph",
    pwc: Pwc,
    kernel: ConvolutionKernel,
    segment_count: int,
    duration: Optional[float] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Filter a piecewise-constant function by convolving it with a kernel and resample it again.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    pwc : Pwc
        The piecewise-constant function :math:`\alpha(t)` to be filtered.
    kernel : ConvolutionKernel
        The node representing the kernel :math:`K(t)`.
    segment_count : int
        The number of segments of the resampled filtered function.
    duration : float or None, optional
        Force the resulting piecewise-constant function to have a certain duration.
        This option is mainly to avoid floating point errors when the total duration is
        too small. Defaults to the sum of segment durations of `pwc`.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The filtered and resampled piecewise-constant function.

    See Also
    --------
    :func:`Graph.convolve_pwc` :
        Create the convolution of a piecewise-constant function with a kernel.
    :func:`Graph.discretize_stf` :
        Create a piecewise-constant function by discretizing a sampleable function.
    :func:`Graph.sinc_convolution_kernel` :
        Create a convolution kernel representing the sinc function.

    Notes
    -----
    The convolution is

    .. math::
        (\alpha * K)(t) \equiv
        \int_{-\infty}^\infty \alpha(\tau) K(t-\tau) \mathrm{d}\tau.

    Convolution in the time domain is equivalent to multiplication in the
    frequency domain, so this function can be viewed as applying a linear
    time-invariant filter (specified via its time domain kernel :math:`K(t)`)
    to :math:`\alpha(t)`.
    """
    total_duration = duration or np.sum(pwc.durations)

    return graph.discretize_stf(
        stf=graph.convolve_pwc(pwc=pwc, kernel=kernel),
        duration=total_duration,
        segment_count=segment_count,
        name=name,
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.node_data import Pwc
from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
)
from boulderopal.graph import Graph


def _validate_addressing(addressing: tuple[int, ...] | int) -> tuple[int, ...]:
    """
    Validate an addressing input and return it as a tuple.
    """
    message = "The ions addressed must be an integer or a tuple of integers."

    if isinstance(addressing, (int, np.integer)):
        return (addressing,)
    try:
        check_argument(
            all(isinstance(ion, (int, np.integer)) for ion in addressing),
            message,
            {"addressing": addressing},
        )
    except TypeError as error:
        raise QctrlArgumentsValueError(message, {"addressing": addressing}) from error

    addressing = tuple(int(k) for k in addressing)
    check_argument(
        len(addressing) == len(set(addressing)),
        "The ions addressed must be unique.",
        {"addressing": addressing},
    )
    return addressing


class Drive:
    """
    A piecewise-constant complex-valued drive.

    Parameters
    ----------
    values : np.ndarray
        The values of the drive at each segment, in units of rad/s.
    addressing : int or tuple[int, ...]
        The indices of the ions addressed by the drive.

    See Also
    --------
    boulderopal.ions.OptimizableDrive :
        Abstract class describing a piecewise-constant optimizable drive.
    boulderopal.ions.ms_simulate :
        Simulate a Mølmer–Sørensen-type operation on a trapped ions system.
    """

    def __init__(self, values: np.ndarray, addressing: int | tuple[int, ...]):
        self.values = ArrayDType.COMPLEX(values, "values", ndim=1)
        self._addressing = _validate_addressing(addressing)

    @property
    def addressing(self) -> tuple[int, ...]:
        """
        The indices of the ions addressed by the drive.
        """
        return self._addressing

    def get_pwc(self, graph: Graph, duration: float) -> Pwc:
        """
        Return a Pwc representation of the drive.
        """
        return graph.pwc_signal(values=self.values, duration=duration)


class OptimizableDrive(ABC):
    """
    Abstract class for optimizable drives. You need to call the concrete classes below
    to create optimizable drives.

    See Also
    --------
    boulderopal.ions.ComplexOptimizableDrive :
        Class describing a piecewise-constant complex-valued optimizable drive.
    boulderopal.ions.RealOptimizableDrive :
        Class describing a piecewise-constant real-valued optimizable drive.
    boulderopal.ions.ms_optimize :
       Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    """

    @abstractmethod
    def get_pwc(self, graph: Graph, duration: float, robust: bool) -> Pwc:
        """
        Return a Pwc representation of the optimizable drive.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Return the name of the drive.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def addressing(self) -> tuple[int, ...]:
        """
        Return the indices of the addressed ions.
        """
        raise NotImplementedError


class ComplexOptimizableDrive(OptimizableDrive):
    """
    A piecewise-constant complex-valued optimizable drive.
    The main function will try to find the optimal values for it.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant drive.
    maximum_rabi_rate : float
        The maximum value that the modulus of the drive can take at each segment,
        in units of rad/s.
    addressing : int or tuple[int, ...]
        The indices of the ions addressed by the drive.
    name : str, optional
        The identifier of the drive.
        Defaults to "drive".

    See Also
    --------
    boulderopal.ions.Drive : Class describing non-optimizable drives.
    boulderopal.ions.RealOptimizableDrive : Class describing optimizable real-valued drives.
    boulderopal.ions.ms_optimize :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    """

    def __init__(
        self,
        count: int,
        maximum_rabi_rate: float,
        addressing: int | tuple[int, ...],
        name: str = "drive",
    ):
        self._count = ScalarDType.INT(count, "count", min_=0)
        self._maximum_rabi_rate = ScalarDType.REAL(
            maximum_rabi_rate, "maximum_rabi_rate", min_=0
        )
        self._addressing = _validate_addressing(addressing)
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def addressing(self) -> tuple[int, ...]:
        return self._addressing

    def get_pwc(self, graph: Graph, duration: float, robust: bool) -> Pwc:
        if not robust:
            return graph.complex_optimizable_pwc_signal(
                segment_count=self._count,
                duration=duration,
                maximum=self._maximum_rabi_rate,
                name=self.name,
            )

        # Create a symmetrized drive signal (Milne et al., Phys. Rev. Applied, 2020).
        free_segment_count = (self._count + 1) // 2
        moduli = graph.optimization_variable(
            count=free_segment_count, lower_bound=0, upper_bound=self._maximum_rabi_rate
        )
        phases = graph.optimization_variable(
            count=free_segment_count,
            lower_bound=0,
            upper_bound=2 * np.pi,
            is_lower_unbounded=True,
            is_upper_unbounded=True,
        )

        if self._count % 2 == 0:
            moduli_reversed = graph.reverse(moduli, [0])
            phases_reversed = graph.reverse(phases, [0])

        else:
            moduli_reversed = graph.reverse(moduli[:-1], [0])
            phases_reversed = graph.reverse(phases[:-1], [0])

        return graph.complex_pwc_signal(
            moduli=graph.concatenate([moduli, moduli_reversed], 0),
            phases=graph.concatenate([phases, 2 * phases[-1] - phases_reversed], 0),
            duration=duration,
            name=self.name,
        )


class RealOptimizableDrive(OptimizableDrive):
    """
    A piecewise-constant real-valued optimizable drive.
    The main function will try to find the optimal values for it.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant drive.
    minimum_rabi_rate : float
        The minimum value that the drive can take at each segment, in units of rad/s.
    maximum_rabi_rate : float
        The maximum value that the drive can take at each segment, in units of rad/s.
    addressing : int or tuple[int, ...]
        The indices of the ions addressed by the drive.
    name : str, optional
        The identifier of the drive.
        Defaults to "drive".

    See Also
    --------
    boulderopal.ions.ComplexOptimizableDrive : Class describing optimizable complex-valued drives.
    boulderopal.ions.Drive : Class describing non-optimizable drives.
    boulderopal.ions.ms_optimize :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    """

    def __init__(
        self,
        count: int,
        minimum_rabi_rate: float,
        maximum_rabi_rate: float,
        addressing: int | tuple[int, ...],
        name: str = "drive",
    ) -> None:
        self._count = ScalarDType.INT(count, "count", min_=0)
        self._minimum_rabi_rate = ScalarDType.REAL(
            minimum_rabi_rate, "minimum_rabi_rate"
        )
        self._maximum_rabi_rate = ScalarDType.REAL(
            maximum_rabi_rate, "maximum_rabi_rate"
        )
        check_argument(
            self._minimum_rabi_rate < self._maximum_rabi_rate,
            "The maximum Rabi rate must be larger than the minimum.",
            {
                "minimum_rabi_rate": minimum_rabi_rate,
                "maximum_rabi_rate": maximum_rabi_rate,
            },
        )
        self._addressing = _validate_addressing(addressing)
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def addressing(self) -> tuple[int, ...]:
        return self._addressing

    def get_pwc(self, graph: Graph, duration: float, robust: bool) -> Pwc:
        if not robust:
            return graph.real_optimizable_pwc_signal(
                segment_count=self._count,
                duration=duration,
                maximum=self._maximum_rabi_rate,
                minimum=self._minimum_rabi_rate,
                name=self.name,
            )

        # Create a symmetrized drive signal (Milne et al., Phys. Rev. Applied, 2020).
        free_segment_count = (self._count + 1) // 2
        values = graph.optimization_variable(
            count=free_segment_count,
            lower_bound=self._minimum_rabi_rate,
            upper_bound=self._maximum_rabi_rate,
        )

        if self._count % 2 == 0:
            values_reversed = graph.reverse(values, [0])

        else:
            values_reversed = graph.reverse(values[:-1], [0])

        return graph.pwc_signal(
            values=graph.concatenate([values, values_reversed], 0),
            duration=duration,
            name=self.name,
        )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.


from __future__ import annotations

from dataclasses import dataclass

from boulderopal._validation.basic import ScalarDType


@dataclass
class Adam:
    """
    Adaptive moment estimation (Adam) optimizer for stochastic optimization.

    For more detail on Adam see
    `Adam <https://en.wikipedia.org/wiki/Stochastic_gradient_descent#Adam>`_ on Wikipedia.

    Parameters
    ----------
    learning_rate : float, optional
        The learning rate for the Adam optimizer.
        If set, must be positive. Defaults to 0.01.

    See Also
    --------
    boulderopal.run_stochastic_optimization :
        Perform gradient-based stochastic optimization of generic real-valued functions.
    """

    learning_rate: float = 0.01

    def __post_init__(self) -> None:
        self.learning_rate = ScalarDType.REAL(
            self.learning_rate, "learning_rate", min_=0
        )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from abc import ABC
from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Any,
    Optional,
)

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
    nullable,
)


@dataclass
class FilterFunction:
    """
    A class to store information about the controls applied in the noise
    reconstruction process, in the form of filter functions.

    The filter function specifies how sensitive a set of controls is to a specific
    noise, at a certain frequency. Filter functions can be calculated using
    the ``execute_graph`` function together with the ``graph.filter_function``
    operator. The output of that operation can then be passed directly to this
    function.

    Parameters
    ----------
    frequencies : np.ndarray
        A 1D array of the frequencies where the filter function is sampled.
        The frequencies must be provided in ascending order.
    inverse_powers : np.ndarray
        A 1D arrays of the values of the filter function at the frequencies
        where it is sampled. Must have the same length as `frequencies`,
        and all its values must be greater or equal to zero.
    uncertainties : np.ndarray or None, optional
        The uncertainties associated with each sampled point of the filter
        function. These values are not used for noise reconstruction.
    """

    frequencies: np.ndarray
    inverse_powers: np.ndarray
    uncertainties: Optional[np.ndarray] = None

    def __post_init__(self) -> None:
        self.frequencies = ArrayDType.REAL(self.frequencies, "frequencies", ndim=1)
        self.inverse_powers = ArrayDType.REAL(
            self.inverse_powers, "inverse_powers", ndim=1, min_=0, min_inclusive=True
        )
        self.uncertainties = nullable(
            ArrayDType.REAL,
            self.uncertainties,
            "uncertainties",
            ndim=1,
            min_=0,
            min_inclusive=True,
        )

        check_argument(
            len(self.frequencies) == len(self.inverse_powers),
            "The frequencies must have the same length as the inverse powers.",
            {"frequencies": self.frequencies, "inverse_powers": self.inverse_powers},
            extras={
                "len(frequencies)": len(self.frequencies),
                "len(inverse_powers)": len(self.inverse_powers),
            },
        )
        check_argument(
            all(np.diff(self.frequencies) > 0),
            "The frequencies must be provided in ascending order.",
            {"frequencies": self.frequencies},
        )
        self._sample_count = len(self.frequencies)

    @property
    def sample_count(self) -> int:
        """
        The number of samples in the filter function.
        """
        return self._sample_count


@dataclass
class NoiseReconstructionMethod(ABC):
    """
    Base class for the noise reconstruction methods.
    """

    method_name: str

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "method_name":
            raise RuntimeError(
                "Mutating the `method_name` of the optimizer is not allowed."
            )
        super().__setattr__(name, value)


@dataclass
class ConvexOptimization(NoiseReconstructionMethod):
    r"""
    Configuration for noise reconstruction with the convex optimization (CVX) method.

    Parameters
    ----------
    power_density_lower_bound : float
        The lower bound for the reconstructed power spectral densities.
        It must be greater than or equal to 0.
    power_density_upper_bound : float
        The upper bound for the reconstructed power spectral densities.
        It must be greater than the `power_density_lower_bound`.
    regularization_hyperparameter : float
        The regularization hyperparameter :math:`\lambda`.

    Notes
    -----
    The CVX method finds the estimation of the power spectral density (PSD) matrix
    :math:`{\mathbf S}` by solving the optimization problem:

    .. math::
        {\mathbf S}_{\mathrm{est}} = \mathrm{argmin}_{\textbf S} (\| F'{\mathbf S} -
        {\mathbf I} \|_2^2 + \lambda \| L_1 {\mathbf S} \|_2^2) ,

    where :math:`F^\prime` is the matrix of weighted filter functions and
    :math:`\| \bullet \|_2` denotes the Euclidean norm and :math:`L_1` is the
    first-order derivative operator defined as

    .. math::
        \begin{align}
            L_1 =
              \begin{bmatrix}
                -1 &      1 &         &    \\
                   & \ddots &  \ddots &     \\
                   &        &      -1 & 1    \\
              \end{bmatrix}_{(K - 1) \times K} .
        \end{align}

    :math:`\lambda` is a positive regularization hyperparameter which determines the
    smoothness of :math:`{\mathbf S}_{\mathrm{est}}`. If you provide uncertainties in
    measurements, this method calculates the uncertainties in estimation using a Monte
    Carlo method.
    """

    power_density_lower_bound: float
    power_density_upper_bound: float
    regularization_hyperparameter: float
    method_name: str = field(default="convex optimization", init=False)

    def __post_init__(self) -> None:
        self.power_density_lower_bound = ScalarDType.REAL(
            self.power_density_lower_bound,
            "power_density_lower_bound",
            min_=0,
            min_inclusive=True,
        )
        self.power_density_upper_bound = ScalarDType.REAL(
            self.power_density_upper_bound, "power_density_upper_bound"
        )
        check_argument(
            self.power_density_lower_bound < self.power_density_upper_bound,
            "The power density lower bound must be less than the upper bound.",
            {
                "self.power_density_lower_bound": self.power_density_lower_bound,
                "self.power_density_upper_bound": self.power_density_upper_bound,
            },
        )

        self.regularization_hyperparameter = ScalarDType.REAL(
            self.regularization_hyperparameter,
            "regularization_hyperparameter",
            min_=0,
            min_inclusive=True,
        )


@dataclass
class SVDEntropyTruncation(NoiseReconstructionMethod):
    r"""
    Configuration for noise reconstruction with the singular value decomposition
    (SVD) method using entropy truncation.

    Parameters
    ----------
    rounding_threshold : float, optional
        The rounding threshold of the entropy, between 0 and 1 (inclusive).
        Defaults to 0.5.

    Notes
    -----
    The singular value decomposition (SVD) method first finds a low rank approximation
    of the matrix of weighted filter functions :math:`F^\prime`:

    .. math::
        F^\prime \approx U \Sigma V ,

    where matrices :math:`U` and :math:`V` satisfy that
    :math:`U^\dagger U = VV^\dagger = \mathbb{I}_{n_{\mathrm{sv}} \times n_{\mathrm{sv}}}`,
    and :math:`\Sigma` is a diagonal matrix of :math:`n_{\mathrm{sv}}` truncated
    singular values, which in the entropy truncation method are determined by
    the entropy of the singular values :math:`E`.

    The entropy truncation method calculates the value :math:`2^E` and rounds the
    value to an integer :math:`n_{\mathrm{sv}}`. When rounding the value
    :math:`2^E`, the floor of :math:`2^E` plus the rounding threshold that you
    chose is taken. Therefore a small value leads to rounding down, while a
    large value leads to rounding up. The :math:`n_{\mathrm{sv}}` is then used
    as the truncation value.

    The SVD method then estimates the noise power spectral density (PSD) :math:`\mathbf S` as:

    .. math::
        {\mathbf S}_{\mathrm{est}} = V^\dagger\Sigma^{-1}U^\dagger{\mathbf I} .

    This method calculates the uncertainties in estimation using error propagation if
    you provide measurement uncertainties.
    """

    rounding_threshold: float = 0.5
    method_name: str = field(default="SVD entropy truncation", init=False)

    def __post_init__(self) -> None:
        self.rounding_threshold = ScalarDType.REAL(
            self.rounding_threshold,
            "rounding_threshold",
            min_=0,
            max_=1,
            min_inclusive=True,
            max_inclusive=True,
        )


@dataclass
class SVDFixedLengthTruncation(NoiseReconstructionMethod):
    r"""
    Configuration for noise reconstruction with the singular value decomposition
    (SVD) method using fixed-length truncation.

    Parameters
    ----------
    singular_value_count : int or None, optional
        The number of singular values to retain. It must be greater or equal to 1.
        Defaults to None, in which case no truncation is performed.

    Notes
    -----
    The singular value decomposition (SVD) method first finds a low rank approximation
    of the matrix of weighted filter functions :math:`F^\prime`:

    .. math::
        F^\prime \approx U \Sigma V ,

    where matrices :math:`U` and :math:`V` satisfy that
    :math:`U^\dagger U = VV^\dagger = \mathbb{I}_{n_{\mathrm{sv}} \times n_{\mathrm{sv}}}`,
    and :math:`\Sigma` is a diagonal matrix of :math:`n_{\mathrm{sv}}` truncated
    singular values, which are determined by the `singular_value_count` that you
    provided.

    The SVD method then estimates the noise power spectral density (PSD) :math:`\mathbf S` as:

    .. math::
        {\mathbf S}_{\mathrm{est}} = V^\dagger\Sigma^{-1}U^\dagger{\mathbf I} .

    This method calculates the uncertainties in estimation using error propagation if
    you provide measurement uncertainties.
    """

    singular_value_count: Optional[int] = None
    method_name: str = field(default="SVD fixed-length truncation", init=False)

    def __post_init__(self) -> None:
        self.singular_value_count = nullable(
            ScalarDType.INT,
            self.singular_value_count,
            "singular_value_count",
            min_=1,
            min_inclusive=True,
        )


ALLOWED_NOISE_RECONSTRUCTION_METHODS = (
    ConvexOptimization,
    SVDEntropyTruncation,
    SVDFixedLengthTruncation,
)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
PWC signal library nodes.
"""
from __future__ import annotations

from enum import Enum
from typing import Optional

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.composite.signal_utils import validate_optimizable_parameter
from boulderopal._nodes.node_data import (
    Pwc,
    Tensor,
)


class SegmentationType(Enum):
    """
    An enumeration of segmentation types for piecewise-constant signals.

    You can use this Enum to specify the segmentation type for graph nodes that
    define a PWC signal containing a constant part.

    Attributes
    ----------
    UNIFORM
        A uniform segmentation distributes the piecewise-constant segments
        uniformly along the signal's duration.
        This segmentation type is more efficient when combining different signals,
        as their segmentations will match.

    MINIMAL
        A minimal segmentation represents constant parts of the piecewise-constant
        function with a minimal number of segments (typically one or two),
        reserving most of the segments for the non-constant parts.
        This segmentation type might be preferred when only a single signal is present
        in the system, as it leads to a more efficient sampling of the non-constant parts.
        However, combining signals with different segmentations can lead to uneven sampling
        and increased computation time due to the increase in number of segments.
    """

    UNIFORM = "UNIFORM"
    MINIMAL = "MINIMAL"


def _get_sample_times(duration: float, segment_count: int) -> np.ndarray:
    """
    Returns an array of `segment_count` equally spaced times between 0 and `duration`.
    Each time is taken at the center of a segment with duration ``dt = duration / segment_count``,
    that is, ``[dt/2, dt + dt/2, 2dt + dt/2, ..., duration - dt/2]``.
    """
    times, time_step = np.linspace(
        0, duration, segment_count, endpoint=False, retstep=True
    )
    times += time_step / 2
    return times


def _parse_segmentation(segmentation: SegmentationType) -> SegmentationType:
    try:
        return SegmentationType(segmentation)

    except ValueError as err:
        raise QctrlArgumentsValueError(
            "Invalid segmentation type.",
            {"segmentation": segmentation},
            extras={"valid types": list(SegmentationType.__members__)},
        ) from err


class PwcSignals:
    """
    Class for the PWC Signal nodes.
    """

    def __init__(self, graph):
        self._graph = graph

    def square_pulse_pwc(
        self,
        duration: float,
        segment_count: int,
        amplitude: float | complex | Tensor,
        start_time: float = 0,
        end_time: Optional[float] = None,
        segmentation: SegmentationType = SegmentationType.UNIFORM,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a square pulse.

        The entire signal lasts from time 0 to the given duration with the
        square pulse being applied from the start time to the end time.

        Parameters
        ----------
        duration : float
            The duration of the signal.
        segment_count : int
            The number of segments in the PWC.
            Only used if the segmentation type is "UNIFORM" .
        amplitude : float or complex or Tensor
            The amplitude of the square pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        start_time : float, optional
            The start time of the square pulse, :math:`t_\mathrm{start}`.
            Defaults to 0.
        end_time : float or None, optional
            The end time of the square pulse, :math:`t_\mathrm{end}`.
            Must be greater than the start time.
            Defaults to the value of the given duration.
        segmentation : :class:`~signals.SegmentationType`
            The type of segmentation for the pulse.
            With a "MINIMAL" segmentation, the returned Pwc has
            between one and three segments, depending on the start time,
            end time, and duration of the signal.
            Defaults to "UNIFORM", in which case the segments are uniformly
            distributed along the signal's duration.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The square pulse.

        See Also
        --------
        :func:`Graph.signals.cosine_pulse_pwc <signals.cosine_pulse_pwc>`
            Create a `Pwc` representing a cosine pulse.
        :func:`Graph.signals.gaussian_pulse_pwc <signals.gaussian_pulse_pwc>`
            Create a `Pwc` representing a Gaussian pulse.
        :func:`Graph.signals.sech_pulse_pwc <signals.sech_pulse_pwc>`
            Create a `Pwc` representing a hyperbolic secant pulse.
        :func:`boulderopal.signals.square_pulse`
            Create a `Signal` object representing a square pulse.

        Notes
        -----
        The square pulse is defined as

        .. math:: \mathop{\mathrm{Square}}(t)
            = A \theta(t-t_\mathrm{start}) \theta(t_\mathrm{end}-t) ,

        where :math:`\theta(t)` is the
        `Heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_.

        Examples
        --------
        Define a square pulse with an optimizable amplitude.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
        ... )
        >>> graph.signals.square_pulse_pwc(
        ...     duration=4.0, amplitude=amplitude, segment_count=100, name="square"
        ... )
        <Pwc: name="square", operation_name="pwc_signal", value_shape=(), batch_shape=()>

        Define a square PWC pulse.

        >>> graph.signals.square_pulse_pwc(
        ...     duration=4.0,
        ...     segment_count=5,
        ...     amplitude=2.5,
        ...     start_time=1.0,
        ...     end_time=3.0,
        ...     name="square",
        ... )
        <Pwc: name="square", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="square")
        >>> result["output"]["square"]
        {
            'durations': array([0.8, 0.8, 0.8, 0.8, 0.8]),
            'values': array([0. , 2.5, 2.5, 2.5, 0. ]),
            'time_dimension': 0
        }

        Define a square PWC pulse with a minimal segmentation.

        >>> graph.signals.square_pulse_pwc(
        ...     duration=4.0,
        ...     segment_count=None,
        ...     amplitude=2.5,
        ...     start_time=1.0,
        ...     end_time=3.0,
        ...     segmentation="MINIMAL",
        ...     name="square",
        ... )
        <Pwc: name="square", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="square")
        >>> result["output"]["square"]
        {
            'durations': array([1., 2., 1.]),
            'values': array([0. , 2.5, 0. ]),
            'time_dimension': 0
        }
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        if end_time is None:
            end_time = duration

        check_argument(
            end_time > start_time,
            "The end time must be greater than the start time.",
            {"start_time": start_time, "end_time": end_time},
        )

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")

        segmentation = _parse_segmentation(segmentation)
        if segmentation is SegmentationType.UNIFORM:
            times = _get_sample_times(duration, segment_count)
            values = amplitude * np.where(
                np.logical_and(times > start_time, times < end_time), 1, 0
            )
            return graph.pwc_signal(values=values, duration=duration, name=name)

        if start_time >= duration or end_time <= 0:
            # In both of these cases the signal is always zero.
            return graph.constant_pwc(constant=0.0, duration=duration, name=name)

        pwcs = []

        if start_time > 0:
            # Add preceding step function.
            pwcs.append(graph.constant_pwc(constant=0.0, duration=start_time))

        pwcs.append(
            graph.constant_pwc(
                constant=amplitude,
                duration=min(end_time, duration) - max(start_time, 0),
            )
        )

        if end_time < duration:
            # Add trailing step function.
            pwcs.append(graph.constant_pwc(constant=0.0, duration=duration - end_time))

        return graph.time_concatenate_pwc(pwcs, name=name)

    def sech_pulse_pwc(
        self,
        duration: float,
        segment_count: int,
        amplitude: float | complex | Tensor,
        width: Optional[float | Tensor] = None,
        center_time: Optional[float | Tensor] = None,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a hyperbolic secant pulse.

        Parameters
        ----------
        duration : float
            The duration of the signal, :math:`T`.
        segment_count : int
            The number of segments in the PWC.
        amplitude : float or complex or Tensor
            The amplitude of the pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        width : float or Tensor or None, optional
            The characteristic time for the hyperbolic secant pulse, :math:`t_\mathrm{pulse}`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to :math:`T/12`,
            giving the pulse a full width at half maximum (FWHM) of :math:`0.22 T`.
        center_time : float or Tensor or None, optional
            The time at which the pulse peaks, :math:`t_\mathrm{peak}`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to :math:`T/2`.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The sampled hyperbolic secant pulse.

        See Also
        --------
        :func:`Graph.signals.cosine_pulse_pwc <signals.cosine_pulse_pwc>`
            Create a `Pwc` representing a cosine pulse.
        :func:`Graph.signals.gaussian_pulse_pwc <signals.gaussian_pulse_pwc>`
            Create a `Pwc` representing a Gaussian pulse.
        :func:`boulderopal.signals.sech_pulse`
            Create a `Signal` object representing a hyperbolic secant pulse.
        :func:`Graph.signals.sech_pulse_stf <signals.sech_pulse_stf>`
            Corresponding operation with `Stf` output.
        :func:`Graph.signals.square_pulse_pwc <signals.square_pulse_pwc>`
            Create a `Pwc` representing a square pulse.

        Notes
        -----
        The hyperbolic secant pulse is defined as

        .. math:: \mathop{\mathrm{Sech}}(t)
            = \frac{A}{\cosh\left((t - t_\mathrm{peak}) / t_\mathrm{pulse} \right)} .

        The FWHM of the pulse is about :math:`2.634 t_\mathrm{pulse}`.

        Examples
        --------
        Define a simple sech PWC pulse.

        >>> graph.signals.sech_pulse_pwc(
        ...     duration=5, segment_count=50, amplitude=1, name="sech"
        ... )
        <Pwc: name="sech", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sech")
        >>> result["output"]["sech"]
        {
            'durations': array([0.1, 0.1, ..., 0.1, 0.1]),
            'values': array([0.00558953, 0.00710565, ..., 0.00710565, 0.00558953]),
            'time_dimension': 0
        }

        Define a displaced sech PWC pulse.

        >>> graph.signals.sech_pulse_pwc(
        ...     duration=3e-6,
        ...     segment_count=60,
        ...     amplitude=20e6,
        ...     width=0.15e-6,
        ...     center_time=1e-6,
        ...     name="sech_displaced",
        ... )
        <Pwc: name="sech_displaced", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sech_displaced")
        >>> result["output"]["sech_displaced"]
        {
            'durations': array([5.e-08, 5.e-08, ..., 5.e-08, 5.e-08]),
            'values': array([6.01374318e+04, 8.39283672e+04, ..., 1.06810547e+02, 7.65331014e+01]),
            'time_dimension': 0
        }

        Define a sech pulse with optimizable parameters.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=10e6, name="amplitude"
        ... )
        >>> width = graph.optimization_variable(
        ...     count=1, lower_bound=0.1e-6, upper_bound=0.5e-6, name="width"
        ... )
        >>> center_time = graph.optimization_variable(
        ...     count=1, lower_bound=1e-6, upper_bound=2e-6, name="center_time"
        ... )
        >>> graph.signals.sech_pulse_pwc(
        ...     duration=3e-6,
        ...     segment_count=32,
        ...     amplitude=amplitude,
        ...     width=width,
        ...     center_time=center_time,
        ...     name="sech_pulse",
        ... )
        <Pwc: name="sech_pulse", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        if width is None:
            width = duration / 12

        if center_time is None:
            center_time = duration / 2

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
        width = validate_optimizable_parameter(graph, width, "width")
        center_time = validate_optimizable_parameter(graph, center_time, "center time")

        stf = amplitude / graph.cosh((graph.identity_stf() - center_time) / width)
        return graph.discretize_stf(stf, duration, segment_count, name=name)

    def linear_ramp_pwc(
        self,
        duration: float,
        segment_count: int,
        end_value: float | complex | Tensor,
        start_value: Optional[float | complex | Tensor] = None,
        start_time: float = 0.0,
        end_time: Optional[float] = None,
        segmentation: SegmentationType = SegmentationType.UNIFORM,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a linear ramp.

        Parameters
        ----------
        duration : float
            The duration of the signal, :math:`T`.
        segment_count : int
            The number of segments in the PWC.
        end_value : float or complex or Tensor
            The value of the ramp at :math:`t = t_\mathrm{end}`, :math:`a_\mathrm{end}`.
            It must either be a scalar or contain a single element.
        start_value : float or complex or Tensor or None, optional
            The value of the ramp at :math:`t = t_\mathrm{start}`, :math:`a_\mathrm{start}`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to :math:`-a_\mathrm{end}`.
        start_time : float, optional
            The time at which the linear ramp starts, :math:`t_\mathrm{start}`.
            Defaults to 0.
        end_time : float or None, optional
            The time at which the linear ramp ends, :math:`t_\mathrm{end}`.
            Defaults to the given duration :math:`T`.
        segmentation : :class:`~signals.SegmentationType`
            The type of segmentation for the signal.
            With a "MINIMAL" segmentation, most of the segments are placed in the
            non-constant parts of the signal.
            Defaults to "UNIFORM", in which case the segments are uniformly
            distributed along the signal's duration.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The sampled linear ramp.

        See Also
        --------
        :func:`boulderopal.signals.linear_ramp`
            Create a `Signal` object representing a linear ramp.
        :func:`Graph.signals.linear_ramp_stf <signals.linear_ramp_stf>`
            Corresponding operation with `Stf` output.
        :func:`Graph.signals.tanh_ramp_pwc <signals.tanh_ramp_pwc>`
            Create a `Pwc` representing a hyperbolic tangent ramp.

        Notes
        -----
        The linear ramp is defined as

        .. math:: \mathop{\mathrm{Linear}}(t) =
            \begin{cases}
                a_\mathrm{start} &\mathrm{if} \quad t < t_\mathrm{start}\\
                a_\mathrm{start} + (a_\mathrm{end} - a_\mathrm{start})
                    \frac{t - t_\mathrm{start}}{t_\mathrm{end} - t_\mathrm{start}}
                    &\mathrm{if} \quad t_\mathrm{start} \le t \le t_\mathrm{end} \\
                a_\mathrm{end} &\mathrm{if} \quad t > t_\mathrm{end}
            \end{cases} .

        Examples
        --------
        Define a linear PWC ramp.

        >>> graph.signals.linear_ramp_pwc(
        ...     duration=2.0, segment_count=5, end_value=1.5, start_value=0.5, name="linear_ramp"
        ... )
        <Pwc: name="linear_ramp", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="linear_ramp")
        >>> result["output"]["linear_ramp"]
        {
            'durations': array([0.4, 0.4, 0.4, 0.4, 0.4]),
            'values': array([0.6, 0.8, 1. , 1.2, 1.4]),
            'time_dimension': 0
        }

        Define a linear ramp with start and end times.

        >>> graph.signals.linear_ramp_pwc(
        ...     duration=4,
        ...     segment_count=8,
        ...     end_value=2,
        ...     start_time=1,
        ...     end_time=3,
        ...     name="linear",
        ... )
        <Pwc: name="linear", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="linear")
        >>> result["output"]["linear"]
        {
            'durations': array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]),
            'values': array([-2. , -2. , -1.5, -0.5,  0.5,  1.5,  2. ,  2. ]),
            'time_dimension': 0
        }

        Define a linear ramp with minimal segmentation.

        >>> graph.signals.linear_ramp_pwc(
        ...     duration=4,
        ...     segment_count=6,
        ...     end_value=2,
        ...     start_time=1,
        ...     end_time=3,
        ...     segmentation="MINIMAL",
        ...     name="linear",
        ... )
        <Pwc: name="linear", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="linear")
        >>> result["output"]["linear"]
        {
            'durations': array([1. , 0.5, 0.5, 0.5, 0.5, 1. ]),
            'values': array([-2. , -1.5, -0.5,  0.5,  1.5,  2. ]),
            'time_dimension': 0
        }

        Define a linear ramp with an optimizable slope around 0.

        >>> duration = 4.0
        >>> slope = graph.optimization_variable(
        ...     count=1, lower_bound=-30, upper_bound=30, name="slope"
        ... )
        >>> end_value = slope * duration / 2
        >>> graph.signals.linear_ramp_pwc(
        ...     duration=duration, segment_count=64, end_value=end_value, name="linear_ramp"
        ... )
        <Pwc: name="linear_ramp", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        if start_value is None:
            start_value = -end_value

        if end_time is None:
            end_time = duration

        check_argument(
            start_time < end_time,
            "The end time of the pulse must be greater than the start time.",
            {"end_time": end_time, "start_time": start_time},
        )

        end_value = validate_optimizable_parameter(graph, end_value, "end value")
        start_value = validate_optimizable_parameter(graph, start_value, "start value")

        segmentation = _parse_segmentation(segmentation)
        if segmentation is SegmentationType.UNIFORM:
            times = _get_sample_times(duration, segment_count)
            values = np.clip((times - start_time) / (end_time - start_time), 0, 1)
            values = (end_value - start_value) * values + start_value
            return graph.pwc_signal(values=values, duration=duration, name=name)

        slope = (end_value - start_value) / (end_time - start_time)

        if start_time <= 0.0 and end_time >= duration:
            # No flat parts inside of the PWC.
            stf = start_value + slope * (graph.identity_stf() - start_time)
            return graph.discretize_stf(stf, duration, segment_count, name=name)

        if end_time <= 0.0 or start_time >= duration:
            # The whole ramp falls outside of the PWC.
            return graph.discretize_stf(
                graph.constant_stf(0.0), duration, segment_count, name=name
            )

        pre_step_pwc = None
        post_step_pwc = None

        if start_time > 0.0:
            # Preceding step function.
            pre_step_pwc = graph.constant_pwc(constant=start_value, duration=start_time)
            segment_count -= 1

        if end_time < duration:
            # Trailing step function.
            post_step_pwc = graph.constant_pwc(
                constant=end_value, duration=duration - end_time
            )
            segment_count -= 1

        # Ramp part.
        stf = start_value + slope * (graph.identity_stf() - min(start_time, 0.0))
        ramp_duration = min(end_time, duration) - max(start_time, 0.0)
        ramp_pwc = graph.discretize_stf(stf, ramp_duration, segment_count)

        pwcs = [ramp_pwc]
        if pre_step_pwc is not None:
            pwcs.insert(0, pre_step_pwc)
        if post_step_pwc is not None:
            pwcs.append(post_step_pwc)

        return graph.time_concatenate_pwc(pwcs, name=name)

    def tanh_ramp_pwc(
        self,
        duration: float,
        segment_count: int,
        end_value: float | complex | Tensor,
        start_value: Optional[float | complex | Tensor] = None,
        ramp_duration: Optional[float | Tensor] = None,
        center_time: Optional[float | Tensor] = None,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a hyperbolic tangent ramp.

        Parameters
        ----------
        duration : float
            The duration of the signal, :math:`T`.
        segment_count : int
            The number of segments in the PWC.
        end_value : float or complex or Tensor
            The asymptotic value of the ramp towards :math:`t \to +\infty`, :math:`a_+`.
            It must either be a scalar or contain a single element.
        start_value : float or complex or Tensor or None, optional
            The asymptotic value of the ramp towards :math:`t \to -\infty`, :math:`a_-`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to minus `end_value`.
        ramp_duration : float or Tensor or None, optional
            The characteristic time for the hyperbolic tangent ramp, :math:`t_\mathrm{ramp}`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to :math:`T/6`.
        center_time : float or Tensor or None, optional
            The time at which the ramp has its greatest slope, :math:`t_0`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to :math:`T/2`.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The sampled hyperbolic tangent ramp.

        See Also
        --------
        :func:`Graph.signals.linear_ramp_pwc <signals.linear_ramp_pwc>`
            Create a `Pwc` representing a linear ramp.
        :func:`boulderopal.signals.tanh_ramp`
            Create a `Signal` object representing a hyperbolic tangent ramp.
        :func:`Graph.signals.tanh_ramp_stf <signals.tanh_ramp_stf>`
            Corresponding operation with `Stf` output.
        :func:`Graph.tanh`
            Calculate the element-wise hyperbolic tangent of an object.

        Notes
        -----
        The hyperbolic tangent ramp is defined as

        .. math:: \mathop{\mathrm{Tanh}}(t)
            = \frac{a_+ + a_-}{2}
                + \frac{a_+ - a_-}{2} \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

        where the function's asymptotic values :math:`a_\pm` are defined by:

        .. math::  a_\pm := \lim_{t\to\pm\infty} \mathop{\mathrm{Tanh}}(t) ,

        and :math:`t_0` is related to :math:`t_\mathrm{ramp}` by:

        .. math::
            \left.\frac{{\rm d}\mathop{\mathrm{Tanh}}(t)}{{\rm d}t}\right|_{t=t_0}
                = \frac{ (a_+ - a_-)}{2 t_\mathrm{ramp}} .

        Note that if :math:`t_0` is close to the edges of the PWC,
        for example :math:`t_0 \lesssim 2 t_\mathrm{ramp}`,
        then the first and last values of the PWC will differ from the expected asymptotic values.

        With the default values of `start_value` (:math:`a_-`),
        `ramp_duration` (:math:`t_\mathrm{ramp}`), and `center_time` (:math:`t_0`),
        the ramp expression simplifies to

        .. math:: \mathop{\mathrm{Tanh}}(t) = A \tanh\left( \frac{t - T/2}{T/6} \right),

        where :math:`A = a_+` is the end value (the start value is then :math:`-A`).
        This defines a symmetric ramp (around :math:`(T/2, 0)`)
        between :math:`-0.995 A` (at :math:`t=0`) and :math:`0.995 A` (at :math:`t=T`).

        Examples
        --------
        Define a simple tanh PWC ramp.

        >>> graph.signals.tanh_ramp_pwc(
        ...     duration=5.0, segment_count=50, end_value=1, name="tanh_ramp"
        ... )
        <Pwc: name="tanh_ramp", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="tanh_ramp")
        >>> result["output"]["tanh_ramp"]
        {
            'durations': array([0.1, 0.1, ..., 0.1, 0.1]),
            'values': array([-0.99442601, -0.99291942, ..., 0.99291942,  0.99442601]),
            'time_dimension': 0
        }

        Define a flat-top pulse from two hyperbolic tangent ramps.

        >>> ramp = graph.signals.tanh_ramp_pwc(
        ...     duration=3,
        ...     segment_count=60,
        ...     end_value=1,
        ...     ramp_duration=0.25,
        ...     center_time=0.5,
        ... )
        >>> flat_top_pulse = 0.5 * (ramp + graph.time_reverse_pwc(ramp))
        >>> flat_top_pulse.name="flat_top_pulse"
        >>> result = bo.execute_graph(graph=graph, output_node_names="flat_top_pulse")
        >>> result["output"]["flat_top_pulse"]
        {
            'durations': array([0.05, 0.05, ..., 0.05, 0.05]),
            'values': array([0.02188127, 0.03229546, ..., 0.03229546, 0.02188127]),
            'time_dimension': 0
        }

        Define a hyperbolic tangent ramp with optimizable parameters.

        >>> end_value = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=3e6, name="end_value"
        ... )
        >>> ramp_duration = graph.optimization_variable(
        ...     count=1, lower_bound=0.1e-6, upper_bound=0.3e-6, name="ramp_duration"
        ... )
        >>> center_time = graph.optimization_variable(
        ...     count=1, lower_bound=0.25e-6, upper_bound=0.75e-6, name="center_time"
        ... )
        >>> graph.signals.tanh_ramp_pwc(
        ...     duration=1e-6,
        ...     segment_count=32,
        ...     end_value=end_value,
        ...     start_value=0.0,
        ...     ramp_duration=ramp_duration,
        ...     center_time=center_time,
        ...     name="tanh_ramp",
        ... )
        <Pwc: name="tanh_ramp", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        if start_value is None:
            start_value = -end_value

        if ramp_duration is None:
            ramp_duration = duration / 6

        if center_time is None:
            center_time = duration / 2

        end_value = validate_optimizable_parameter(graph, end_value, "end value")
        start_value = validate_optimizable_parameter(graph, start_value, "start value")
        ramp_duration = validate_optimizable_parameter(
            graph, ramp_duration, "ramp duration"
        )
        center_time = validate_optimizable_parameter(graph, center_time, "center time")

        stf = start_value + 0.5 * (end_value - start_value) * (
            1 + graph.tanh((graph.identity_stf() - center_time) / ramp_duration)
        )
        return graph.discretize_stf(stf, duration, segment_count, name=name)

    def gaussian_pulse_pwc(
        self,
        duration: float,
        segment_count: int,
        amplitude: float | complex | Tensor,
        width: Optional[float | Tensor] = None,
        center_time: Optional[float] = None,
        drag: Optional[float | Tensor] = None,
        flat_duration: Optional[float] = None,
        segmentation: SegmentationType = SegmentationType.UNIFORM,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a Gaussian pulse.

        Parameters
        ----------
        duration : float
            The duration of the signal, :math:`T`.
        segment_count : int
            The number of segments in the PWC.
            Must be at least four.
        amplitude : float or complex or Tensor
            The amplitude of the Gaussian pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        width : float or Tensor or None, optional
            The standard deviation of the Gaussian pulse, :math:`\sigma`.
            It must either be a scalar or contain a single element.
            Defaults to :math:`T/10` or :math:`(T-t_\mathrm{flat})/10` if `flat_duration` is passed.
        center_time : float or None, optional
            The center of the Gaussian pulse, :math:`t_0`.
            Defaults to half of the given value of the duration, :math:`T/2`.
        drag : float or Tensor or None, optional
            The DRAG parameter, :math:`\beta`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to no DRAG correction.
        flat_duration : float or None, optional
            The amount of time to remain constant after the peak of the Gaussian,
            :math:`t_\mathrm{flat}`.
            If passed, it must be positive and less than the duration.
            Defaults to None, in which case no constant part is added to the Gaussian pulse.
        segmentation : :class:`~signals.SegmentationType`
            The type of segmentation for the signal.
            With a "MINIMAL" segmentation, most of the segments are placed in the
            non-constant parts of the signal.
            Defaults to "UNIFORM", in which case the segments are uniformly
            distributed along the signal's duration.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The sampled Gaussian pulse.
            If no flat duration is passed then the pulse is evenly sampled between :math:`0` and
            :math:`T`. If one is passed, the flat part of the pulse is described by one or two segments
            (depending on the values of `center_time` and `segment_count`), and the rest of the pulse is
            sampled with the remaining segments.

        See Also
        --------
        :func:`Graph.signals.cosine_pulse_pwc <signals.cosine_pulse_pwc>`
            Create a `Pwc` representing a cosine pulse.
        :func:`Graph.signals.gaussian_pulse_stf <signals.gaussian_pulse_stf>`
            Corresponding operation with `Stf` output.
        :func:`Graph.signals.sech_pulse_pwc <signals.sech_pulse_pwc>`
            Create a `Pwc` representing a hyperbolic secant pulse.
        :func:`Graph.signals.square_pulse_pwc <signals.square_pulse_pwc>`
            Create a `Pwc` representing a square pulse.

        Notes
        -----
        The Gaussian pulse is defined as

        .. math:: \mathop{\mathrm{Gaussian}}(t) =
            \begin{cases}
                A \left(1-\frac{i\beta (t-t_1)}{\sigma^2}\right)
                \exp \left(- \frac{(t-t_1)^2}{2\sigma^2} \right)
                    &\mathrm{if} \quad t < t_1=t_0- t_\mathrm{flat}/2\\
                A
                    &\mathrm{if} \quad t_0-t_\mathrm{flat}/2 \le t < t_0+t_\mathrm{flat}/2 \\
                A \left(1-\frac{i\beta (t-t_2)}{\sigma^2}\right)
                \exp \left(- \frac{(t-t_2)^2}{2\sigma^2} \right)
                    &\mathrm{if} \quad t > t_2=t_0+t_\mathrm{flat}/2
            \end{cases} .

        If the flat duration is zero (the default setting), this reduces to

        .. math:: \mathop{\mathrm{Gaussian}}(t) =
            A \left(1-\frac{i\beta (t-t_0)}{\sigma^2}\right)
            \exp \left(- \frac{(t-t_0)^2}{2\sigma^2} \right) .

        Examples
        --------
        Define a Gaussian PWC pulse.

        >>> graph.signals.gaussian_pulse_pwc(
        ...     duration=3.0,
        ...     segment_count=100,
        ...     amplitude=1.0,
        ...     name="gaussian",
        ... )
        <Pwc: name="gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="gaussian")
        >>> result["output"]["gaussian"]
        {
            'durations': array([0.03, 0.03, ..., 0.03, 0.03]),
            'values': array([4.77913973e-06, 7.80106730e-06, ..., 7.80106730e-06, 4.77913973e-06]),
            'time_dimension': 0
        }

        Define a flat-top Gaussian PWC pulse with a DRAG correction with minimal segmentation.

        >>> graph.signals.gaussian_pulse_pwc(
        ...     duration=3.0,
        ...     segment_count=100,
        ...     amplitude=1.0,
        ...     width=0.2,
        ...     center_time=1.5,
        ...     drag=0.1,
        ...     flat_duration=0.2,
        ...     segmentation="MINIMAL",
        ...     name="gaussian_drag",
        ... )
        <Pwc: name="gaussian_drag", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="gaussian_drag")
        >>> result["output"]["gaussian_drag"]
        {
            'durations': array([0.02857143, 0.02857143, ..., 0.02857143, 0.02857143]),
            'values': array([3.76551948e-11+1.30448351e-10j, 1.00289674e-10+3.40268532e-10j, ...,
                             1.00289593e-10-3.40268262e-10j, 3.76551637e-11-1.30448246e-10j]),
            'time_dimension': 0
        }

        Define a Gaussian pulse with optimizable parameters.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
        ... )
        >>> width = graph.optimization_variable(
        ...     count=1, lower_bound=0.1, upper_bound=2., name="width"
        ... )
        >>> drag = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=1., name="drag"
        ... )
        >>> graph.signals.gaussian_pulse_pwc(
        ...     duration=3.0,
        ...     segment_count=100,
        ...     amplitude=amplitude,
        ...     width=width,
        ...     drag=drag,
        ...     name="gaussian",
        ... )
        <Pwc: name="gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        """
        # pylint:enable=line-too-long
        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        if center_time is None:
            center_time = 0.5 * duration

        if width is None:
            if flat_duration is None:
                width = duration / 10
            else:
                width = (duration - flat_duration) / 10

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
        width = validate_optimizable_parameter(graph, width, "width")

        def create_gaussian(center_parameter):
            assert width is not None
            shifted_time = graph.identity_stf() - center_parameter
            gaussian = amplitude * graph.exp(-(shifted_time**2) / (2 * width**2))
            if drag is None:
                return gaussian
            return gaussian * (1.0 - 1j * drag * shifted_time / (width**2))

        if drag is not None:
            drag = validate_optimizable_parameter(graph, drag, "drag")

        if flat_duration is None:
            return graph.discretize_stf(
                create_gaussian(center_time), duration, segment_count, name=name
            )

        check_argument(
            0.0 < flat_duration < duration,
            "The flat duration must be positive and less than the duration.",
            {"flat_duration": flat_duration},
            extras={"duration": duration},
        )

        check_argument(
            segment_count > 3,
            "The number of segments must be at least 4.",
            {"segment_count": segment_count},
        )

        # Time at which first Gaussian ends.
        flat_segment_start = center_time - 0.5 * flat_duration
        # Time at which second Gaussian starts.
        flat_segment_end = center_time + 0.5 * flat_duration

        segmentation = _parse_segmentation(segmentation)
        if segmentation is SegmentationType.UNIFORM:
            times = _get_sample_times(duration, segment_count)

            left_ramp_times = times[np.where(times < flat_segment_start)]
            values = []
            if len(left_ramp_times) > 0:
                values.append(
                    graph.sample_stf(
                        create_gaussian(flat_segment_start), left_ramp_times
                    )
                )

            flat_times_count = np.sum(
                np.logical_and(times >= flat_segment_start, times <= flat_segment_end)
            )
            if flat_times_count > 0:
                values.append(amplitude * np.ones(flat_times_count))

            right_ramp_times = times[np.where(times > flat_segment_end)]
            if len(right_ramp_times) > 0:
                values.append(
                    graph.sample_stf(
                        create_gaussian(flat_segment_end), right_ramp_times
                    )
                )

            return graph.pwc_signal(
                values=graph.concatenate(values, axis=0), duration=duration, name=name
            )

        if flat_segment_start >= duration:
            # In this case since the flat segment starts after the duration, it is not part of
            # the signal.
            return graph.discretize_stf(
                create_gaussian(flat_segment_end), duration, segment_count, name=name
            )

        if flat_segment_end <= 0:
            # In this case since the flat segment finishes before time zero, it is not part of
            # the signal.
            return graph.discretize_stf(
                create_gaussian(flat_segment_end), duration, segment_count, name=name
            )

        if flat_segment_end >= duration:
            # In this case since the flat segment finishes after the duration, the second Gaussian
            # is not part of the signal.
            durations = [flat_segment_start, duration - flat_segment_start]
            segment_counts = [segment_count - 1, 1]
            stfs = [create_gaussian(flat_segment_start), graph.constant_stf(amplitude)]

        elif flat_segment_start <= 0:
            # In this case since the flat segment begins before time zero, the first Gaussian
            # is not part of the signal.
            durations = [flat_segment_end, duration - flat_segment_end]
            segment_counts = [1, segment_count - 1]
            stfs = [graph.constant_stf(amplitude), create_gaussian(0.0)]

        else:
            durations = [flat_segment_start, flat_duration, duration - flat_segment_end]
            segment_counts = _allocate_segment_counts(durations, segment_count)
            stfs = [
                create_gaussian(flat_segment_start),
                graph.constant_stf(amplitude),
                create_gaussian(0.0),
            ]

        pwcs = [
            graph.discretize_stf(*args) for args in zip(stfs, durations, segment_counts)
        ]
        return graph.time_concatenate_pwc(pwcs, name=name)

    def cosine_pulse_pwc(
        self,
        duration: float,
        segment_count: int,
        amplitude: float | complex | Tensor,
        drag: Optional[float | Tensor] = None,
        start_time: float = 0.0,
        end_time: Optional[float] = None,
        flat_duration: Optional[float] = None,
        segmentation: SegmentationType = SegmentationType.UNIFORM,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a cosine pulse.

        Parameters
        ----------
        duration : float
            The duration of the signal, :math:`T`.
        segment_count : int
            The number of segments in the PWC.
            Must be at least six.
        amplitude : float or complex or Tensor
            The amplitude of the pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        drag : float or Tensor or None, optional
            The DRAG parameter, :math:`\beta`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to no DRAG correction.
        start_time : float, optional
            The time at which the cosine pulse starts, :math:`t_\mathrm{start}`.
            Defaults to 0.
        end_time : float or None, optional
            The time at which the cosine pulse ends, :math:`t_\mathrm{end}`.
            Defaults to the given duration :math:`T`.
        flat_duration : float or None, optional
            The amount of time to remain constant after the peak of the cosine,
            :math:`t_\mathrm{flat}`.
            If passed, it must be positive and less than the difference between `end_time` and
            `start_time`.
            Defaults to None, in which case no constant part is added to the cosine pulse.
        segmentation : :class:`~signals.SegmentationType`
            The type of segmentation for the signal.
            With a "MINIMAL" segmentation, most of the segments are placed in the
            non-constant parts of the signal.
            Defaults to "UNIFORM", in which case the segments are uniformly
            distributed along the signal's duration.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The sampled cosine pulse.
            If no flat duration is passed then the pulse is evenly sampled between :math:`0` and
            :math:`T`. If one is passed, the flat part of the pulse is described by one or two segments
            (depending on the value of `segment_count`), and the rest of the pulse is sampled with the
            remaining segments.

        See Also
        --------
        :func:`boulderopal.signals.cosine_pulse`
            Create a `Signal` object representing a cosine pulse.
        :func:`Graph.signals.gaussian_pulse_pwc <signals.gaussian_pulse_pwc>`
            Create a `Pwc` representing a Gaussian pulse.
        :func:`Graph.signals.hann_series_pwc <signals.hann_series_pwc>`
            Create a `Pwc` representing a sum of Hann window functions.
        :func:`Graph.signals.sech_pulse_pwc <signals.sech_pulse_pwc>`
            Create a `Pwc` representing a hyperbolic secant pulse.
        :func:`Graph.signals.sinusoid_pwc <signals.sinusoid_pwc>`
            Create a `Pwc` representing a sinusoidal oscillation.
        :func:`Graph.signals.square_pulse_pwc <signals.square_pulse_pwc>`
            Create a `Pwc` representing a square pulse.
        :func:`Graph.cos`
            Calculate the element-wise cosine of an object.

        Notes
        -----
        The cosine pulse is defined as

        .. math:: \mathop{\mathrm{Cos}}(t) =
            \begin{cases}
            0
            &\mathrm{if} \quad t < t_\mathrm{start} \\
            \frac{A}{2} \left[1+\cos \left(\omega \{t-\tau_-\} \right)
            + i\omega\beta \sin \left(\omega \{t-\tau_-\}\right)\right]
            &\mathrm{if} \quad t_\mathrm{start} \le t < \tau_- \\
            A
            &\mathrm{if} \quad \tau_- \le t \le \tau_+ \\
            \frac{A}{2} \left[1+\cos \left(\omega\{t-\tau_+\}\right)
            + i\omega \beta\sin \left(\omega \{t-\tau_+\}\right)\right]
            &\mathrm{if} \quad \tau_+ < t \le t_\mathrm{end} \\
            0
            &\mathrm{if} \quad t > t_\mathrm{end} \\
            \end{cases},

        where :math:`\omega=2\pi /(t_\mathrm{end}-t_\mathrm{start} - t_\mathrm{flat})`,
        :math:`\tau_\mp` are the start/end times of the flat segment,
        with :math:`\tau_\mp=(t_\mathrm{start}+t_\mathrm{end} \mp t_\mathrm{flat})/2`.

        If the flat duration is zero (the default setting), this reduces to

        .. math:: \mathop{\mathrm{Cos}}(t) =
            \frac{A}{2} \left[1+\cos \left(\omega \{t-\tau\} \right)
            + i\omega\beta \sin \left(\omega \{t-\tau\}\right)\right]
            \theta(t-t_\mathrm{start}) \theta(t_\mathrm{end}-t),

        where now :math:`\omega=2\pi /(t_\mathrm{end}-t_\mathrm{start})`,
        :math:`\tau=(t_\mathrm{start}+t_\mathrm{end})/2`
        and :math:`\theta(t)` is the
        `Heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_.

        Examples
        --------
        Define a cosine PWC pulse.

        >>> graph.signals.cosine_pulse_pwc(
        ...     duration=3.0, segment_count=100, amplitude=1.0, name="cos_pulse"
        ... )
        <Pwc: name="cos_pulse", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="cos_pulse")
        >>> result["output"]["cos_pulse"]
        {
            'durations': array([0.03, 0.03, ..., 0.03, 0.03]),
            'values': array([2.46719817e-04, 2.21901770e-03, ..., 2.21901770e-03, 2.46719817e-04]),
            'time_dimension': 0
        }

        Define a flat-top cosine PWC pulse with a DRAG correction.

        >>> graph.signals.cosine_pulse_pwc(
        ...     duration=3.0,
        ...     segment_count=100,
        ...     amplitude=1.0,
        ...     drag=0.1,
        ...     start_time=1.0,
        ...     end_time=2.0,
        ...     flat_duration=0.3,
        ...     segmentation="MINIMAL",
        ...     name="cos_flat",
        ... )
        <Pwc: name="cos_flat", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="cos_flat")
        >>> result["output"]["cos_flat"]
        {
            'durations': array([1.        , 0.00729167, ..., 0.00729167, 1.        ]),
            'values': array([0.00000000e+00+0.j        , 2.67706262e-04-0.01468429j, ...,
                             2.67706262e-04+0.01468429j, 0.00000000e+00+0.j        ]),
            'time_dimension': 0
        }

        Define a cosine pulse with optimizable parameters.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
        ... )
        >>> drag = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=1., name="drag"
        ... )
        >>> graph.signals.cosine_pulse_pwc(
        ...     duration=3.0,
        ...     segment_count=100,
        ...     amplitude=amplitude,
        ...     drag=drag,
        ...     name="cos_pulse",
        ... )
        <Pwc: name="cos_pulse", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        if end_time is None:
            end_time = duration

        check_argument(
            start_time < end_time,
            "The end time of the pulse must be greater than the start time.",
            {"end_time": end_time, "start_time": start_time},
        )

        if flat_duration is not None:
            check_argument(
                0 < flat_duration < end_time - start_time,
                "The flat duration must be positive and less than the end time of the "
                "pulse minus the start time.",
                {
                    "flat_duration": flat_duration,
                    "start_time": start_time,
                    "end_time": end_time,
                },
                extras={"end_time - start_time": end_time - start_time},
            )

        if start_time >= duration or end_time <= 0:
            # In both of these cases the signal is always zero.
            return graph.pwc_signal(
                duration=duration, values=np.zeros(segment_count), name=name
            )

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")

        if drag is not None:
            drag = validate_optimizable_parameter(graph, drag, "drag")

        segmentation = _parse_segmentation(segmentation)
        if segmentation is SegmentationType.UNIFORM:
            if flat_duration is None:
                flat_duration = 0.0
            pulse_period = end_time - start_time - flat_duration
            angular_freq = 2 * np.pi / pulse_period
            pulse_center = 0.5 * (start_time + end_time)
            flat_part_start = pulse_center - flat_duration * 0.5
            flat_part_end = pulse_center + flat_duration * 0.5

            times = _get_sample_times(duration, segment_count)

            # Create array with the values of the argument of cos/sin in the pulse definition.
            phases = np.where(
                np.logical_and(times > flat_part_start, times < flat_part_end), 0, np.pi
            )
            left_ramp_mask = np.where(
                np.logical_and(times >= start_time, times <= flat_part_start)
            )
            phases[left_ramp_mask] = angular_freq * (
                times[left_ramp_mask] - flat_part_start
            )
            right_ramp_mask = np.where(
                np.logical_and(times >= flat_part_end, times <= end_time)
            )
            phases[right_ramp_mask] = angular_freq * (
                times[right_ramp_mask] - flat_part_end
            )

            # Calculate pulse values.
            pulse = 1 + np.cos(phases)
            if drag is not None:
                pulse = pulse + angular_freq * drag * 1j * np.sin(phases)

            return graph.pwc_signal(
                duration=duration, values=0.5 * amplitude * pulse, name=name
            )

        def create_pulse(peak_time, omega, duration_, segment_count_):
            # Creates a PWC with segment_count_ segments representing
            #   A/2 [ 1 + cos(ω (t-tp)) + i ω β sin(ω (t - tp)) ]
            # between 0 and duration_,
            # where A is the amplitude, ω is omega, tp is peak_time,
            # and β is the DRAG parameter.

            shifted_times = _get_sample_times(duration_, segment_count_) - peak_time
            pulse = 1 + np.cos(omega * shifted_times)
            if drag is not None:
                pulse = pulse + omega * drag * 1j * np.sin(omega * shifted_times)

            return graph.pwc_signal(duration=duration_, values=0.5 * amplitude * pulse)

        pwcs = []

        pulse_segment_count = (
            segment_count - (start_time > 0.0) - (end_time < duration)
        )  # The number of segments for the cosine pulse.

        if start_time > 0.0:
            # Add preceding step function.
            pwcs.append(graph.pwc_signal(duration=start_time, values=np.array([0.0])))

        if flat_duration is None:
            pulse_period = end_time - start_time  # The period of the cosine pulse.
            angular_freq = 2 * np.pi / pulse_period
            # The peak is 0.5 * pulse_period after the
            # start of the pulse plus start_time if start_time < 0.
            pwcs.append(
                create_pulse(
                    0.5 * pulse_period + min(start_time, 0),
                    angular_freq,
                    min(duration, end_time) - max(start_time, 0),
                    pulse_segment_count,
                )
            )
        else:
            check_argument(
                segment_count > 5,
                "The number of segments must be at least 6.",
                {"segment_count": segment_count},
            )

            # The period of the cosine pulse.
            pulse_period = end_time - start_time - flat_duration
            angular_freq = 2 * np.pi / pulse_period
            pulse_center = 0.5 * (start_time + end_time)
            flat_segment_start = pulse_center - flat_duration * 0.5
            flat_segment_end = pulse_center + flat_duration * 0.5

            durations = np.array(
                [
                    min(flat_segment_start, duration) - max(start_time, 0),
                    min(flat_segment_end, duration) - max(flat_segment_start, 0),
                    min(end_time, duration) - max(flat_segment_end, 0),
                ]
            )
            durations = durations * (durations > 0.0)

            if flat_segment_start <= 0.0 and flat_segment_end >= duration:
                # In this case the flat segment is the entire signal.
                segment_counts = [0, pulse_segment_count, 0]
            else:
                segment_counts = _allocate_segment_counts(
                    durations, pulse_segment_count
                )

            if flat_segment_start > 0.0:
                pwcs.append(
                    create_pulse(
                        0.5 * pulse_period + min(start_time, 0),
                        angular_freq,
                        durations[0],
                        segment_counts[0],
                    )
                )

            if flat_segment_start < duration and flat_segment_end > 0.0:
                pwcs.append(
                    graph.discretize_stf(
                        graph.constant_stf(amplitude), durations[1], segment_counts[1]
                    )
                )

            if flat_segment_end < duration:
                # The peak is at the start of the pulse, unless flat_segment_end < 0 in
                # which case it's flat_segment_end.
                pwcs.append(
                    create_pulse(
                        min(flat_segment_end, 0),
                        angular_freq,
                        durations[2],
                        segment_counts[2],
                    )
                )

        if end_time < duration:
            # Add trailing step function.
            pwcs.append(
                graph.pwc_signal(duration=duration - end_time, values=np.array([0.0]))
            )

        return graph.time_concatenate_pwc(pwcs, name=name)

    def sinusoid_pwc(
        self,
        duration: float,
        segment_count: int,
        amplitude: float | complex | Tensor,
        angular_frequency: float | Tensor,
        phase: float | Tensor = 0.0,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a sinusoidal oscillation.

        Parameters
        ----------
        duration : float
            The duration of the signal, :math:`T`.
        segment_count : int
            The number of segments in the PWC.
        amplitude : float or complex or Tensor
            The amplitude of the oscillation, :math:`A`.
            It must either be a scalar or contain a single element.
        angular_frequency : float or Tensor
            The angular frequency of the oscillation, :math:`\omega`.
            It must either be a scalar or contain a single element.
        phase : float or Tensor, optional
            The phase of the oscillation, :math:`\phi`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The sampled sinusoid.

        See Also
        --------
        :func:`Graph.signals.cosine_pulse_pwc <signals.cosine_pulse_pwc>`
            Create a `Pwc` representing a cosine pulse.
        :func:`Graph.signals.hann_series_pwc <signals.hann_series_pwc>`
            Create a `Pwc` representing a sum of Hann window functions.
        :func:`boulderopal.signals.sinusoid`
            Create a `Signal` object representing a sinusoidal oscillation.
        :func:`Graph.signals.sinusoid_stf <signals.sinusoid_stf>`
            Corresponding operation with `Stf` output.
        :func:`Graph.sin`
            Calculate the element-wise sine of an object.

        Notes
        -----
        The sinusoid is defined as

        .. math:: \mathop{\mathrm{Sinusoid}}(t) = A \sin \left( \omega t + \phi \right) .

        Examples
        --------
        Define a PWC oscillation.

        >>> graph.signals.sinusoid_pwc(
        ...     duration=5.0,
        ...     segment_count=100,
        ...     amplitude=1.0,
        ...     angular_frequency=np.pi,
        ...     phase=np.pi/2.0,
        ...     name="oscillation"
        ... )
        <Pwc: name="oscillation", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="oscillation")
        >>> result["output"]["oscillation"]
        {
            'durations': array([0.05, 0.05, ..., 0.05, 0.05]),
            'values': array([ 0.99691733,  0.97236992,  ..., -0.97236992, -0.99691733]),
            'time_dimension': 0
        }

        Define a sinusoid with optimizable parameters.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=4e3, name="amplitude"
        ... )
        >>> angular_frequency = graph.optimization_variable(
        ...     count=1, lower_bound=5e6, upper_bound=20e6, name="angular_frequency"
        ... )
        >>> phase = graph.optimization_variable(
        ...     count=1,
        ...     lower_bound=0,
        ...     upper_bound=2*np.pi,
        ...     is_lower_unbounded=True,
        ...     is_upper_unbounded=True,
        ...     name="phase",
        ... )
        >>> graph.signals.sinusoid_pwc(
        ...     duration=3e-6,
        ...     segment_count=100,
        ...     amplitude=amplitude,
        ...     angular_frequency=angular_frequency,
        ...     phase=phase,
        ...     name="oscillation"
        ... )
        <Pwc: name="oscillation", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
        angular_frequency = validate_optimizable_parameter(
            graph, angular_frequency, "angular frequency"
        )
        phase = validate_optimizable_parameter(graph, phase, "phase")

        stf = amplitude * graph.sin(angular_frequency * graph.identity_stf() + phase)
        return graph.discretize_stf(stf, duration, segment_count, name=name)

    def hann_series_pwc(
        self,
        duration: float,
        segment_count: int,
        coefficients: np.ndarray | Tensor,
        name: Optional[str] = None,
    ) -> Pwc:
        # pylint: disable=line-too-long
        r"""
        Create a `Pwc` representing a sum of Hann window functions.

        The piecewise-constant function is sampled from Hann functions that start and end at zero.

        Parameters
        ----------
        duration : float
            The duration of the signal, :math:`T`.
        segment_count : int
            The number of segments in the PWC.
        coefficients : np.ndarray or Tensor
            The coefficients for the different Hann window functions, :math:`c_n`.
            It must be a 1D array or Tensor and it can't contain more than `segment_count` elements.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The sampled Hann window functions series.

        See Also
        --------
        :func:`Graph.signals.cosine_pulse_pwc <signals.cosine_pulse_pwc>`
            Create a `Pwc` representing a cosine pulse.
        :func:`Graph.signals.sinusoid_pwc <signals.sinusoid_pwc>`
            Create a `Pwc` representing a sinusoidal oscillation.
        :func:`boulderopal.signals.hann_series`
            Create a `Signal` object representing a sum of Hann window functions.
        :func:`Graph.signals.hann_series_stf <signals.hann_series_stf>`
            Corresponding operation with `Stf` output.

        Notes
        -----
        The series is defined as

        .. math:: \mathop{\mathrm{Hann}}(t)
            = \sum_{n=1}^N c_n \sin^2 \left( \frac{\pi n t}{T} \right) ,

        where :math:`N` is the number of coefficients.

        Examples
        --------
        Define a simple Hann series.

        >>> graph.signals.hann_series_pwc(
        ...     duration=5.0,
        ...     segment_count=50,
        ...     coefficients=np.array([0.5, 1, 0.25]),
        ...     name="hann_series",
        ... )
        <Pwc: name="hann_series", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="hann_series")
        >>> result["output"]["hann_series"]
        {
            'durations': array([0.1, 0.1, ..., 0.1, 0.1]),
            'values': array([0.00665006, 0.05899895, ..., 0.05899895, 0.00665006]),
            'time_dimension': 0
        }

        Define a Hann series with optimizable coefficients.

        >>> coefficients = graph.optimization_variable(
        ...     count=8, lower_bound=-3.5e6, upper_bound=3.5e6, name="coefficients"
        ... )
        >>> graph.signals.hann_series_pwc(
        ...     duration=2.0e-6,
        ...     segment_count=128,
        ...     coefficients=coefficients,
        ...     name="hann_series",
        ... )
        <Pwc: name="hann_series", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            duration > 0.0, "The duration must be positive.", {"duration": duration}
        )

        check_argument(
            len(coefficients.shape) == 1,
            "The coefficients must be in a 1D array or Tensor.",
            {"coefficients": coefficients},
            extras={"coefficients.shape": coefficients.shape},
        )

        check_argument(
            coefficients.shape[0] <= segment_count,
            "There can't be more coefficients than segments.",
            {"coefficients": coefficients, "segment_count": segment_count},
            extras={"coefficients.shape": coefficients.shape},
        )

        # Define scaled times π t / T to sample the function.
        scaled_times = _get_sample_times(duration, segment_count) * np.pi / duration

        # Calculate function values.
        nss = np.arange(1, coefficients.shape[0] + 1)
        values = graph.sum(
            coefficients * np.sin(nss * scaled_times[:, None]) ** 2, axis=1
        )

        return graph.pwc_signal(duration=duration, values=values, name=name)


def _allocate_segment_counts(durations: list | np.ndarray, segment_count: int) -> list:
    """
    Allocate the number of segments between two non-flat segments and one flat segment.
    The flat segment is assumed to lie in the middle of the other two.
    The number of segments each non-flat part has is in proportion to their duration of time.
    The non-flat durations are at indexes 0 and 2 of durations, durations[0] and durations[2].
    durations[1] can only be zero if one of durations[0] or durations[2] are also zero
    (e.g. [2,0,3] and [0,0,0] are not valid inputs).
    """
    # In the symmetric case (durations[0] = durations[2]), the flat part gets two segments
    # if segment_count is even and one otherwise.
    if np.isclose(durations[0], durations[2]):
        pulse_counts = [(segment_count - 1) // 2] * 2
        pulse_counts.insert(1, segment_count - 2 * pulse_counts[0])
        return pulse_counts

    flat_segments = int(durations[1] > 0.0)  # Number of flat segments.
    # Note that if flat_segments is 0, either durations[0] or durations[2] must be 0.
    if np.isclose(durations[0], 0.0):
        return [0, flat_segments, segment_count - flat_segments]
    if np.isclose(durations[2], 0.0):
        return [segment_count - flat_segments, flat_segments, 0]

    # In this case we know flat_segments=1 and all pulse_counts should be at least 1.
    non_flat_duration = durations[0] + durations[2]
    pulse_counts = [
        int((segment_count - 1) * durations[index] / non_flat_duration)
        for index in [0, 2]
    ]
    # Make sure the pulses each get at least one segment even if the duration is very small.
    pulse_counts = [max(pulse_count, 1) for pulse_count in pulse_counts]
    pulse_counts.insert(1, 1)

    # If there are too many or few segments, we update the largest segment count.
    pulse_counts[np.argmax(pulse_counts)] += segment_count - sum(pulse_counts)

    return pulse_counts



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Classes associated to the closed-loop optimizers.
"""

from __future__ import annotations

from abc import ABC
from dataclasses import (
    dataclass,
    field,
)
from enum import Enum
from typing import (
    Any,
    Optional,
)

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
    nullable,
    validate_enum,
)


class BoundType(Enum):
    """
    Boundary type for parameter bounds.

    Attributes
    ----------
    NON_PERIODIC
        Non-periodic boundaries. This means the optimizer will clip the parameter values
        to remain within the bounds any time it takes a step outside of the bounds.

    PERIODIC
        Periodic boundaries. This means the optimizer will modulo parameter values
        back within the bounds any time it takes a step outside of the bounds.

    See also
    --------
    boulderopal.closed_loop.Bounds :
        A box constraint with which you can define the (inclusive) bounds and their type
        for each optimizable parameter in your optimization.
    """

    NON_PERIODIC = "NON_PERIODIC"
    PERIODIC = "PERIODIC"


@dataclass
class Bounds:
    """
    A box constraint with which you can define the (inclusive) bounds and their type
    for each optimizable parameter in your optimization.

    Parameters
    ----------
    values : np.ndarray
        The per-parameter bounds on the test points.
        The bounds must be a NumPy array of shape ``(parameter_count, 2)`` where the trailing
        axis are the bounds for each parameter (with the lower bound first, followed by the upper
        bound).
    bound_type : BoundType or list[BoundType], optional
        The boundary type for the bounds. You can provide a single BoundType to be used
        for all parameters, or a list with a BoundType for each parameter.
        Defaults to non-periodic bounds.
    """

    values: np.ndarray
    bound_type: BoundType | list[BoundType] | list[str] = BoundType.NON_PERIODIC

    def __post_init__(self) -> None:
        self.values = _validate_bounds(self.values)
        if isinstance(self.bound_type, list):
            check_argument(
                len(self.bound_type) == self.values.shape[0],
                "If you provide a list of bound types, it must have as many items"
                " as parameters",
                {"values": self.values, "bound_type": self.bound_type},
            )
            self.bound_type = [
                validate_enum(BoundType, bound_type) for bound_type in self.bound_type
            ]
        else:
            bound_type = validate_enum(BoundType, self.bound_type)
            self.bound_type = [bound_type] * self.values.shape[0]


@dataclass
class ClosedLoopOptimizer(ABC):  # pylint: disable=too-few-public-methods
    """
    Abstract class for optimizers used in closed-loop optimization.

    To create an optimizer, use one of the concrete classes.

    See Also
    --------
    boulderopal.closed_loop.Cmaes :
        Class describing the covariance matrix adaptation evolution strategy (CMA-ES) optimizer.
    boulderopal.closed_loop.GaussianProcess :
        Class describing the Gaussian-process-based optimizer.
    boulderopal.closed_loop.NeuralNetwork :
        Class describing the neural-network-based optimizer.
    boulderopal.closed_loop.SimulatedAnnealing :
        Class describing the simulated annealing optimizer.
    """

    bounds: Bounds
    method_name: str

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "method_name":
            raise RuntimeError(
                "Mutating the `method_name` of the optimizer is not allowed."
            )
        super().__setattr__(name, value)


@dataclass
class Cmaes(ClosedLoopOptimizer):
    r"""
    The covariance matrix adaptation evolution strategy (CMA-ES) optimizer.

    Parameters
    ----------
    bounds : Bounds
        The bounds on the test points.
    initial_mean : np.ndarray or None, optional
        The per-parameter initial mean for the multivariate normal distribution.
        Defaults to None, in which case a random value inside the bounds is used for each parameter.
        If set, each parameter's mean must be within its corresponding bounds.
    initial_step_size : float or None, optional
        The initial step size for the multivariate normal distribution from which new test
        points are sampled.
        Defaults to one.
    population_size : int or None, optional
        The population size of the test candidates. It is recommended to use a population size of
        at least :math:`P = 4 + \lfloor 3 \times \log N \rfloor`, where :math:`N` is the number of
        optimizable parameters, :math:`\log` is the natural logarithm, and :math:`\lfloor x \rfloor`
        is the floor function.
        Defaults to :math:`P`.
    seed : int or None, optional
        Seed for the random number generator used in the optimizer.
        If set, must be a non-negative integer.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    The CMA-ES optimizer uses a multivariate normal distribution to generate new test points.
    From an `initial_mean` and `initial_step_size`, this distribution is continually
    updated using an evolutionary strategy, with each update depending on the previous values and
    the current set of results.
    New test points are sampled from the distribution until convergence is reached.

    For more detail on CMA-ES see `CMA-ES <https://en.wikipedia.org/wiki/CMA-ES>`_ on Wikipedia.
    """

    bounds: Bounds
    initial_mean: Optional[np.ndarray] = None
    initial_step_size: Optional[float] = None
    population_size: Optional[int] = None
    seed: Optional[int] = None
    method_name: str = field(default="CMA-ES", init=False)

    def __post_init__(self) -> None:
        self.initial_mean = nullable(
            ArrayDType.REAL, self.initial_mean, "initial_mean", ndim=1
        )
        self.population_size = nullable(
            ScalarDType.INT, self.population_size, "population_size", min_=0
        )
        self.seed = nullable(
            ScalarDType.INT, self.seed, "seed", min_=0, min_inclusive=True
        )

        if self.initial_mean is not None:
            check_argument(
                len(self.initial_mean) == len(self.bounds.values),
                "The initial mean and the bounds must have the same length.",
                {"initial_mean": self.initial_mean},
                extras={"bounds.values": self.bounds.values},
            )
            for mean, bound in zip(self.initial_mean, self.bounds.values):
                check_argument(
                    bound[0] <= mean <= bound[1],
                    "The initial mean must be within the bounds.",
                    {"initial_mean": self.initial_mean},
                    extras={"bounds.values": self.bounds.values},
                )


@dataclass
class GaussianProcess(ClosedLoopOptimizer):
    r"""
    The Gaussian process optimizer.

    Parameters
    ----------
    bounds : Bounds
        The bounds on the test points.
    length_scale_bounds : np.ndarray or None, optional
        The per-parameter length scale bounds on the test points.
        The bounds must be a NumPy array of shape ``(parameter_count, 2)`` where the trailing
        axis are the bounds for each parameter (with the lower bound first, followed by the upper
        bound).
        If not specified,  :py:obj:`~closed_loop.optimize` will pick a value derived from the
        `bounds` by picking orders of magnitudes below/above the sidelength for each box axis.
    seed : int or None, optional
        Seed for the random number generator used in the optimizer.
        If set, must be a non-negative integer.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    The Gaussian process is defined by the kernel

    .. math::
        k({\mathbf x}_j, {\mathbf x}_k)
            = \exp \left(-\frac{1}{2} ( {\mathbf x}_j - {\mathbf x}_k )^\top
                \Sigma^{-2} ( {\mathbf x}_j - {\mathbf x}_k )\right) ,

    where :math:`{\mathbf x}_j` is an :math:`n`-dimensional vector representing the
    :math:`j`-th test point, :math:`\Sigma= {\rm diag}(l_1, \cdots, l_n)`
    is an :math:`n \times n` diagonal matrix, and :math:`\{ l_j \}` are the length scales.
    The length scales are tuned while training the model, within the bounds set by the
    `length_scale_bounds` parameter. Roughly speaking, the amount a parameter needs to change
    to impact the optimization cost should lie within the length scale bounds.

    It's recommended to provide non-zero `cost_uncertainty` to :py:obj:`~closed_loop.optimize`
    when using this optimizer, otherwise you might encounter a numerical error when the optimizer
    tries to fit the kernel with your input data. If the error persists, try increasing the
    `cost_uncertainty` value or decreasing the minimum length scale bound. However, such numerical
    error is also an indication that your data might not be suitable to be modelled by a
    Gaussian process, and in that case, consider using a different closed-loop optimizer.

    For more detail on Gaussian processes see
    `Gaussian process <https://en.wikipedia.org/wiki/Gaussian_process>`_ on Wikipedia.
    """

    bounds: Bounds
    length_scale_bounds: Optional[np.ndarray] = None
    seed: Optional[int] = None
    method_name: str = field(default="Gaussian process", init=False)

    def __post_init__(self) -> None:
        if self.length_scale_bounds is not None:
            self.length_scale_bounds = _validate_bounds(
                self.length_scale_bounds, "length scale bounds"
            )

        self.seed = nullable(
            ScalarDType.INT, self.seed, "seed", min_=0, min_inclusive=True
        )

        if self.length_scale_bounds is not None:
            check_argument(
                len(self.length_scale_bounds) == len(self.bounds.values),
                "The length scale bounds and the bounds must have the same length.",
                {"length_scale_bounds": self.length_scale_bounds},
                extras={"bounds.values": self.bounds.values},
            )


@dataclass
class NeuralNetwork(ClosedLoopOptimizer):
    """
    The neural network optimizer.

    Parameters
    ----------
    bounds : Bounds
        The bounds on the test points.
    seed : int or None, optional
        Seed for the random number generator used in the optimizer.
        If set, must be a non-negative integer.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    The neural network optimizer builds and trains a neural network to fit the cost landscape with
    the data it receives. Then a set of test points are returned, which minimize the neural
    network's fitted cost landscape. A gradient based optimizer is used to minimize this landscape,
    with the points starting from different random initial values.

    This method is recommended when you can provide a large amount of data about your system.

    The network architecture used by this optimizer is chosen for its good performance on a variety
    of quantum control tasks.

    For best results, you should pass an array of `initial_parameters` evenly sampled
    over the whole parameter space.
    """

    bounds: Bounds
    seed: Optional[int] = None
    method_name: str = field(default="neural network", init=False)

    def __post_init__(self) -> None:
        self.seed = nullable(
            ScalarDType.INT, self.seed, "seed", min_=0, min_inclusive=True
        )


@dataclass
class SimulatedAnnealing(ClosedLoopOptimizer):
    r"""
    The simulated annealing optimizer.

    Parameters
    ----------
    bounds : Bounds
        The bounds on the test points.
    temperatures : np.ndarray
        The array of initial per-parameter annealing temperatures :math:`T_0` used to generate
        new test points.
        Higher temperatures correspond to higher exploration.
        The per-parameter adjustments from the current test point are sampled from Cauchy
        distributions with scales given by temperatures.
        The temperatures are currently implemented to decay such that each temperature
        at the k-th step is set according to :math:`T_k=\frac{T_0}{1+k}`.
        All temperatures must be positive.
    temperature_cost : float
        The parameter for controlling the optimizer’s greediness.
        A high cost temperature allows the optimizer to explore test points which may not
        immediately improve the cost. A higher level of exploration can be helpful for
        more difficult optimization problems. The cost temperature is set to decay
        according to the same schedule as the temperatures.
        Must be positive.
    seed : int or None, optional
        Seed for the random number generator used in the optimizer.
        If set, must be a non-negative integer.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    This simulated annealing performs a temperature based random walk within the parameter `bounds`.
    The new test points are sampled from a distribution whose variance is given by the current
    `temperatures`. Higher `temperatures` translate to high levels of exploration, which can be
    useful for non-convex optimization. The `temperature_cost` parameter can be set independently of
    the `temperatures`, and controls the overall greediness of each update. A high
    `temperature_cost` allows the optimizer to accept updates which do not immediately improve the
    cost. Both `temperatures` and the `temperature_cost` automatically decay between iterations.

    For more information on this method see
    `simulated annealing <https://en.wikipedia.org/wiki/Simulated_annealing>`_ on Wikipedia.
    """

    bounds: Bounds
    temperatures: np.ndarray
    temperature_cost: float
    seed: Optional[int] = None
    method_name: str = field(default="simulated annealing", init=False)

    def __post_init__(self) -> None:
        self.temperatures = ArrayDType.REAL(
            self.temperatures, "temperatures", min_=0, ndim=1
        )
        self.temperature_cost = ScalarDType.REAL(
            self.temperature_cost, "temperature_cost", min_=0
        )
        self.seed = nullable(
            ScalarDType.INT, self.seed, "seed", min_=0, min_inclusive=True
        )

        check_argument(
            len(self.temperatures) == len(self.bounds.values),
            "The temperatures and the bounds must have the same length.",
            {"temperatures": self.temperatures},
            extras={"bounds.values": self.bounds.values},
        )


def _validate_bounds(bounds: np.ndarray, name: str = "bounds") -> np.ndarray:
    """
    Validate input bounds.
    """
    bounds = ArrayDType.REAL(bounds, name, ndim=2)
    check_argument(
        bounds.shape[1] == 2,
        f"The {name} must be a 2D array with two components in the second axis.",
        {name: bounds},
        extras={f"{name}.shape": bounds.shape},
    )
    check_argument(
        all(bounds[:, 1] > bounds[:, 0]),
        "The upper bound (second component) must be greater than the lower bound "
        f"(first component) for each element in the {name} array.",
        {name: bounds},
    )
    return bounds


ALLOWED_OPTIMIZERS = (Cmaes, NeuralNetwork, GaussianProcess, SimulatedAnnealing)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Namespaces for functions, nodes, and classes.
"""

from functools import partial

NAMESPACE_ATTR = "namespace_name"
COMPOSITE_ATTR = "composite_node"


def composite_node(categories: list):
    """
    Define a decorator that:
        - Sets a new attribute `composite_node` to True.
        - Sets a new attribute `categories` to a given value.
    """
    assert isinstance(categories, list), "The categories must be a list."

    def _set_node_information(obj, categories):
        setattr(obj, "categories", categories)
        setattr(obj, COMPOSITE_ATTR, True)
        return obj

    return partial(_set_node_information, categories=categories)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from boulderopal._configuration.activity_monitor import show_activity_monitor
from boulderopal._configuration.cloud import (
    get_result,
    group_requests,
    request_machines,
    set_organization,
    show_machine_status,
)
from boulderopal._configuration.verbosity import set_verbosity

__all__ = [
    "get_result",
    "group_requests",
    "request_machines",
    "set_organization",
    "set_verbosity",
    "show_activity_monitor",
    "show_machine_status",
]
